from __future__ import absolute_import
import os
import io
from irods.models import DataObject, Collection
from irods.manager import Manager
from irods.message import (
    iRODSMessage, FileOpenRequest, ObjCopyRequest, StringStringMap, DataObjInfo, ModDataObjMeta,
    DataObjChksumRequest, DataObjChksumResponse, RErrorStack)
import irods.exception as ex
from irods.api_number import api_number
from irods.data_object import (
    iRODSDataObject, iRODSDataObjectFileRaw, chunks, irods_dirname, irods_basename)
import irods.keywords as kw
import irods.parallel as parallel
import six
import ast
import json
import logging


MAXIMUM_SINGLE_THREADED_TRANSFER_SIZE = 32 * ( 1024 ** 2)

DEFAULT_NUMBER_OF_THREADS = 0   # Defaults for reasonable number of threads -- optimized to be
                                # performant but allow no more worker threads than available CPUs.
                                # Setting this to 1 disables automatic use of parallel transfer.
DEFAULT_QUEUE_DEPTH = 32


class Server_Checksum_Warning(Exception):
    """Error from iRODS server indicating some replica checksums are missing or incorrect."""
    def __init__(self,json_response):
        """Initialize the exception object with a checksum field from the server response message."""
        super(Server_Checksum_Warning,self).__init__()
        self.response = json.loads(json_response)


class DataObjectManager(Manager):

    READ_BUFFER_SIZE = 1024 * io.DEFAULT_BUFFER_SIZE
    WRITE_BUFFER_SIZE = 1024 * io.DEFAULT_BUFFER_SIZE

    # Data object open flags (independent of client os)
    O_RDONLY = 0
    O_WRONLY = 1
    O_RDWR = 2
    O_APPEND = 1024
    O_CREAT = 64
    O_EXCL = 128
    O_TRUNC = 512


    def should_parallelize_transfer( self,
                                     num_threads = 0,
                                     obj_sz = 1+MAXIMUM_SINGLE_THREADED_TRANSFER_SIZE,
                                     server_version_hint = ()):
        # Allow an environment variable to override the detection of the server version.
        # Example: $ export IRODS_VERSION_OVERRIDE="4,2,9" ;  python -m irods.parallel ...
        # ---
        # Delete the following line on resolution of https://github.com/irods/irods/issues/5932 :
        if getattr(self.sess,'ticket__',None) is not None: return False
        server_version = ( ast.literal_eval(os.environ.get('IRODS_VERSION_OVERRIDE', '()' )) or server_version_hint or 
                           self.server_version )
        if num_threads == 1 or ( server_version < parallel.MINIMUM_SERVER_VERSION ):
            return False
        if getattr(obj_sz,'seek',None) :
            pos = obj_sz.tell()
            size = obj_sz.seek(0,os.SEEK_END)
            if not isinstance(size,six.integer_types):
                size = obj_sz.tell()
            obj_sz.seek(pos,os.SEEK_SET)
        else:
            size = obj_sz
            assert (size > -1)
        return size > MAXIMUM_SINGLE_THREADED_TRANSFER_SIZE


    def _download(self, obj, local_path, num_threads, **options):

        if os.path.isdir(local_path):
            local_file = os.path.join(local_path, irods_basename(obj))
        else:
            local_file = local_path

        # Check for force flag if local_file exists
        if os.path.exists(local_file) and kw.FORCE_FLAG_KW not in options:
            raise ex.OVERWRITE_WITHOUT_FORCE_FLAG

        with open(local_file, 'wb') as f, self.open(obj, 'r', **options) as o:

            if self.should_parallelize_transfer (num_threads, o):
                f.close()
                if not self.parallel_get( (obj,o), local_path, num_threads = num_threads,
                                          target_resource_name = options.get(kw.RESC_NAME_KW,'')):
                    raise RuntimeError("parallel get failed")
            else:
                for chunk in chunks(o, self.READ_BUFFER_SIZE):
                    f.write(chunk)


    def get(self, path, local_path = None, num_threads = DEFAULT_NUMBER_OF_THREADS, **options):
        parent = self.sess.collections.get(irods_dirname(path))

        # TODO: optimize
        if local_path:
            self._download(path, local_path, num_threads = num_threads, **options)

        query = self.sess.query(DataObject)\
            .filter(DataObject.name == irods_basename(path))\
            .filter(DataObject.collection_id == parent.id)\
            .add_keyword(kw.ZONE_KW, path.split('/')[1])

        if hasattr(self.sess,'ticket__'):
            query = query.filter(Collection.id != 0) # a no-op, but necessary because CAT_SQL_ERR results if the ticket
                                                     # is for a DataObject and we don't explicitly join to Collection

        results = query.all() # get up to max_rows replicas
        if len(results) <= 0:
            raise ex.DataObjectDoesNotExist()
        return iRODSDataObject(self, parent, results)


    def put(self, local_path, irods_path, return_data_object = False, num_threads = DEFAULT_NUMBER_OF_THREADS, **options):

        if irods_path.endswith('/'):
            obj = irods_path + os.path.basename(local_path)
        else:
            obj = irods_path

        with open(local_path, 'rb') as f, self.open(obj, 'w', **options) as o:

            if self.should_parallelize_transfer (num_threads, f):
                f.close()
                if not self.parallel_put( local_path, (obj,o), num_threads = num_threads,
                                          target_resource_name = options.get(kw.RESC_NAME_KW,'') or
                                                                 options.get(kw.DEST_RESC_NAME_KW,'')):
                    raise RuntimeError("parallel put failed")
            else:
                # Set operation type to trigger acPostProcForPut
                if kw.OPR_TYPE_KW not in options:
                    options[kw.OPR_TYPE_KW] = 1 # PUT_OPR
                for chunk in chunks(f, self.WRITE_BUFFER_SIZE):
                    o.write(chunk)

        if kw.ALL_KW in options:
            options[kw.UPDATE_REPL_KW] = ''
            self.replicate(obj, **options)

        if return_data_object:
            return self.get(obj)


    def chksum(self, path, **options):
        """
        See: https://github.com/irods/irods/blob/4-2-stable/lib/api/include/dataObjChksum.h
        for a list of applicable irods.keywords options.
        """
        r_error_stack = options.pop('r_error',None)
        message_body = DataObjChksumRequest(path, **options)
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_CHKSUM_AN'])
        checksum = ""
        msg_retn = []
        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            try:
                response = conn.recv(return_message = msg_retn)
            except ex.CHECK_VERIFICATION_RESULTS as exc:
                # We'll get a response in the client to help qualify or elaborate on the error thrown.
                if msg_retn: response = msg_retn[0]
                logging.warning("Exception checksumming data object %r - %r",path,exc)
            if 'response' in locals():
                try:
                    results = response.get_main_message(DataObjChksumResponse, r_error = r_error_stack)
                    checksum = results.myStr.strip()
                    if checksum[0] in ( '[','{' ):  # in iRODS 4.2.11 and later, myStr is in JSON format.
                        exc = Server_Checksum_Warning( checksum )
                        if not r_error_stack:
                            r_error_stack.fill(exc.response)
                        raise exc
                except iRODSMessage.ResponseNotParseable:
                    # response.msg is None when VERIFY_CHKSUM_KW is used
                    pass
        return checksum


    def parallel_get(self,
                     data_or_path_ ,
                     file_ ,
                     async_ = False,
                     num_threads = 0,
                     target_resource_name = '',
                     progressQueue = False):

        return parallel.io_main( self.sess, data_or_path_, parallel.Oper.GET | (parallel.Oper.NONBLOCKING if async_ else 0), file_,
                                 num_threads = num_threads, target_resource_name = target_resource_name,
                                 queueLength = (DEFAULT_QUEUE_DEPTH if progressQueue else 0))

    def parallel_put(self,
                     file_ ,
                     data_or_path_ ,
                     async_ = False,
                     num_threads = 0,
                     target_resource_name = '',
                     progressQueue = False):

        return parallel.io_main( self.sess, data_or_path_, parallel.Oper.PUT | (parallel.Oper.NONBLOCKING if async_ else 0), file_,
                                 num_threads = num_threads, target_resource_name = target_resource_name,
                                 queueLength = (DEFAULT_QUEUE_DEPTH if progressQueue else 0))


    def create(self, path, resource=None, force=False, **options):
        options[kw.DATA_TYPE_KW] = 'generic'

        if resource:
            options[kw.DEST_RESC_NAME_KW] = resource
        else:
            # Use client-side default resource if available
            try:
                options[kw.DEST_RESC_NAME_KW] = self.sess.default_resource
            except AttributeError:
                pass

        if force:
            options[kw.FORCE_FLAG_KW] = ''

        message_body = FileOpenRequest(
            objPath=path,
            createMode=0o644,
            openFlags=0,
            offset=0,
            dataSize=-1,
            numThreads=self.sess.numThreads,
            oprType=0,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_CREATE_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()
            desc = response.int_info
            conn.close_file(desc)

        return self.get(path)


    def open_with_FileRaw(self, *arg, **kw_options):
        holder = []
        handle = self.open(*arg,_raw_fd_holder=holder,**kw_options)
        return (handle, holder[-1])

    def open(self, path, mode, create = True, finalize_on_close = True, **options):
        _raw_fd_holder =  options.get('_raw_fd_holder',[])
        if kw.DEST_RESC_NAME_KW not in options:
            # Use client-side default resource if available
            try:
                options[kw.DEST_RESC_NAME_KW] = self.sess.default_resource
            except AttributeError:
                pass
        createFlag = self.O_CREAT if create else 0
        flags, seek_to_end = {
            'r': (self.O_RDONLY, False),
            'r+': (self.O_RDWR, False),
            'w': (self.O_WRONLY | createFlag | self.O_TRUNC, False),
            'w+': (self.O_RDWR | createFlag | self.O_TRUNC, False),
            'a': (self.O_WRONLY | createFlag, True),
            'a+': (self.O_RDWR | createFlag, True),
        }[mode]
        # TODO: Use seek_to_end

        try:
            oprType = options[kw.OPR_TYPE_KW]
        except KeyError:
            oprType = 0

        message_body = FileOpenRequest(
            objPath=path,
            createMode=0,
            openFlags=flags,
            offset=0,
            dataSize=-1,
            numThreads=self.sess.numThreads,
            oprType=oprType,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_OPEN_AN'])

        conn = self.sess.pool.get_connection()
        conn.send(message)
        desc = conn.recv().int_info

        raw = iRODSDataObjectFileRaw(conn, desc, finalize_on_close = finalize_on_close, **options)
        (_raw_fd_holder).append(raw)
        return io.BufferedRandom(raw)


    def unlink(self, path, force=False, **options):
        if force:
            options[kw.FORCE_FLAG_KW] = ''

        try:
            oprType = options[kw.OPR_TYPE_KW]
        except KeyError:
            oprType = 0

        message_body = FileOpenRequest(
            objPath=path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=-1,
            numThreads=self.sess.numThreads,
            oprType=oprType,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_UNLINK_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()


    def unregister(self, path, **options):
        # https://github.com/irods/irods/blob/4.2.1/lib/api/include/dataObjInpOut.h#L190
        options[kw.OPR_TYPE_KW] = 26

        self.unlink(path, **options)


    def exists(self, path):
        try:
            self.get(path)
        except ex.DoesNotExist:
            return False
        return True


    def move(self, src_path, dest_path):
        # check if dest is a collection
        # if so append filename to it
        if self.sess.collections.exists(dest_path):
            filename = src_path.rsplit('/', 1)[1]
            target_path = dest_path + '/' + filename
        else:
            target_path = dest_path

        src = FileOpenRequest(
            objPath=src_path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=0,
            numThreads=self.sess.numThreads,
            oprType=11,   # RENAME_DATA_OBJ
            KeyValPair_PI=StringStringMap(),
        )
        dest = FileOpenRequest(
            objPath=target_path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=0,
            numThreads=self.sess.numThreads,
            oprType=11,   # RENAME_DATA_OBJ
            KeyValPair_PI=StringStringMap(),
        )
        message_body = ObjCopyRequest(
            srcDataObjInp_PI=src,
            destDataObjInp_PI=dest
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_RENAME_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()


    def copy(self, src_path, dest_path, **options):
        # check if dest is a collection
        # if so append filename to it
        if self.sess.collections.exists(dest_path):
            filename = src_path.rsplit('/', 1)[1]
            target_path = dest_path + '/' + filename
        else:
            target_path = dest_path

        src = FileOpenRequest(
            objPath=src_path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=0,
            numThreads=self.sess.numThreads,
            oprType=10,   # COPY_SRC
            KeyValPair_PI=StringStringMap(),
        )
        dest = FileOpenRequest(
            objPath=target_path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=0,
            numThreads=self.sess.numThreads,
            oprType=9,   # COPY_DEST
            KeyValPair_PI=StringStringMap(options),
        )
        message_body = ObjCopyRequest(
            srcDataObjInp_PI=src,
            destDataObjInp_PI=dest
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_COPY_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()


    def truncate(self, path, size, **options):
        message_body = FileOpenRequest(
            objPath=path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=size,
            numThreads=self.sess.numThreads,
            oprType=0,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_TRUNCATE_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()


    def replicate(self, path, resource=None, **options):
        if resource:
            options[kw.DEST_RESC_NAME_KW] = resource

        message_body = FileOpenRequest(
            objPath=path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=-1,
            numThreads=self.sess.numThreads,
            oprType=6,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_REPL_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()


    def register(self, file_path, obj_path, **options):
        options[kw.FILE_PATH_KW] = file_path

        message_body = FileOpenRequest(
            objPath=obj_path,
            createMode=0,
            openFlags=0,
            offset=0,
            dataSize=0,
            numThreads=self.sess.numThreads,
            oprType=0,
            KeyValPair_PI=StringStringMap(options),
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['PHY_PATH_REG_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()

    def modDataObjMeta(self, data_obj_info, meta_dict, **options):
        if "rescHier" not in data_obj_info and "rescName" not in data_obj_info and "replNum" not in data_obj_info:
            meta_dict["all"] = ""
            
        message_body = ModDataObjMeta(
            dataObjInfo=DataObjInfo(
                objPath=data_obj_info["objPath"],
                rescName=data_obj_info.get("rescName", ""),
                rescHier=data_obj_info.get("rescHier", ""),
                dataType="",
                dataSize=0,
                chksum="",
                version="",
                filePath="",
                dataOwnerName="",
                dataOwnerZone="",
                replNum=data_obj_info.get("replNum", 0),
                replStatus=0,
                statusString="",
                dataId=0,
                collId=0,
                dataMapId=0,
                flags=0,
                dataComments="",
                dataMode="",
                dataExpiry="",
                dataCreate="",
                dataModify="",
                dataAccess="",
                dataAccessInx=0,
                writeFlag=0,
                destRescName="",
                backupRescName="",
                subPath="",
                specColl=0,
                regUid=0,
                otherFlags=0,
                KeyValPair_PI=StringStringMap(options),
                in_pdmo="",
                next=0,
                rescId=0
                ),
            regParam=StringStringMap(meta_dict)
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['MOD_DATA_OBJ_META_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(message)
            response = conn.recv()



# if you're copying these from the docs, you might find the following regex helpful:
# s/\(\w\+\)\s\+\(-\d\+\)/class \1(SystemException):\r    code = \2/g


from __future__ import absolute_import
import six
import numbers


class PycommandsException(Exception):
    pass


class NetworkException(PycommandsException):
    pass


class DoesNotExist(PycommandsException):
    pass


class DataObjectDoesNotExist(DoesNotExist):
    pass


class CollectionDoesNotExist(DoesNotExist):
    pass


class ZoneDoesNotExist(DoesNotExist):
    pass


class UserDoesNotExist(DoesNotExist):
    pass


class UserGroupDoesNotExist(DoesNotExist):
    pass


class ResourceDoesNotExist(DoesNotExist):
    pass


class OperationNotSupported(PycommandsException):
    pass


class QueryException(PycommandsException):
    pass


class NoResultFound(QueryException):
    pass


class MultipleResultsFound(QueryException):
    pass


class iRODSExceptionMeta(type):
    codes = {}

    def __init__(self, name, bases, attrs):
        if 'code' in attrs:
            iRODSExceptionMeta.codes[attrs['code']] = self


class iRODSException(six.with_metaclass(iRODSExceptionMeta, Exception)):
    pass


def nominal_code( the_code, THRESHOLD = 1000 ):
    nominal = []
    c = rounded_code( the_code , nominal_int_repo = nominal )
    negated = -abs(nominal[0])
    return c if (negated <= -abs(THRESHOLD)) else negated  # produce a negative for nonzero integer input

def rounded_code( the_code , nominal_int_repo = () ):
    nom_err = None
    try:
        if isinstance(the_code,type) and \
           issubclass(the_code, iRODSException): the_code = getattr( the_code, 'code', the_code )
        if isinstance(the_code,str):
            nom_err = globals()[the_code].code
            return nom_err
        elif isinstance(the_code,numbers.Integral):
            nom_err = the_code
            return 1000 * ((-abs(the_code) - 1) // 1000 + 1)
        else:
            message = 'Supplied code {the_code!r} must be integer or string'.format(**locals())
            raise RuntimeError(message)
    finally:
        if nom_err is not None and isinstance(nominal_int_repo,list):
            nominal_int_repo[:] = [nom_err]


def get_exception_class_by_code(code, name_only=False):
    rounded = rounded_code (code)  # rounded up to -1000 if code <= -1000
    cls = iRODSExceptionMeta.codes.get( rounded )
    return cls if not name_only \
      else (cls.__name__ if cls is not None else 'Unknown_iRODS_error')


def get_exception_by_code(code, message = None):
    exc_class = iRODSExceptionMeta.codes[ rounded_code( code ) ]
    exc_instance = exc_class( message )
    exc_instance.code = code
    return exc_instance


class UnknowniRODSError(iRODSException):
    code = 0  # covers rounded_code (errcode) if 0 > errcode > -1000


class SystemException(iRODSException):
    pass


class SYS_SOCK_OPEN_ERR(SystemException):
    code = -1000


class SYS_SOCK_BIND_ERR(SystemException):
    code = -2000


class SYS_SOCK_ACCEPT_ERR(SystemException):
    code = -3000


class SYS_HEADER_READ_LEN_ERR(SystemException):
    code = -4000


class SYS_HEADER_WRITE_LEN_ERR(SystemException):
    code = -5000


class SYS_HEADER_TPYE_LEN_ERR(SystemException):
    code = -6000


class SYS_CAUGHT_SIGNAL(SystemException):
    code = -7000


class SYS_GETSTARTUP_PACK_ERR(SystemException):
    code = -8000


class SYS_EXCEED_CONNECT_CNT(SystemException):
    code = -9000


class SYS_USER_NOT_ALLOWED_TO_CONN(SystemException):
    code = -10000


class SYS_READ_MSG_BODY_INPUT_ERR(SystemException):
    code = -11000


class SYS_UNMATCHED_API_NUM(SystemException):
    code = -12000


class SYS_NO_API_PRIV(SystemException):
    code = -13000


class SYS_API_INPUT_ERR(SystemException):
    code = -14000


class SYS_PACK_INSTRUCT_FORMAT_ERR(SystemException):
    code = -15000


class SYS_MALLOC_ERR(SystemException):
    code = -16000


class SYS_GET_HOSTNAME_ERR(SystemException):
    code = -17000


class SYS_OUT_OF_FILE_DESC(SystemException):
    code = -18000


class SYS_FILE_DESC_OUT_OF_RANGE(SystemException):
    code = -19000


class SYS_UNRECOGNIZED_REMOTE_FLAG(SystemException):
    code = -20000


class SYS_INVALID_SERVER_HOST(SystemException):
    code = -21000


class SYS_SVR_TO_SVR_CONNECT_FAILED(SystemException):
    code = -22000


class SYS_BAD_FILE_DESCRIPTOR(SystemException):
    code = -23000


class SYS_INTERNAL_NULL_INPUT_ERR(SystemException):
    code = -24000


class SYS_CONFIG_FILE_ERR(SystemException):
    code = -25000


class SYS_INVALID_ZONE_NAME(SystemException):
    code = -26000


class SYS_COPY_LEN_ERR(SystemException):
    code = -27000


class SYS_PORT_COOKIE_ERR(SystemException):
    code = -28000


class SYS_KEY_VAL_TABLE_ERR(SystemException):
    code = -29000


class SYS_INVALID_RESC_TYPE(SystemException):
    code = -30000


class SYS_INVALID_FILE_PATH(SystemException):
    code = -31000


class SYS_INVALID_RESC_INPUT(SystemException):
    code = -32000


class SYS_INVALID_PORTAL_OPR(SystemException):
    code = -33000


class SYS_PARA_OPR_NO_SUPPORT(SystemException):
    code = -34000


class SYS_INVALID_OPR_TYPE(SystemException):
    code = -35000


class SYS_NO_PATH_PERMISSION(SystemException):
    code = -36000


class SYS_NO_ICAT_SERVER_ERR(SystemException):
    code = -37000


class SYS_AGENT_INIT_ERR(SystemException):
    code = -38000


class SYS_PROXYUSER_NO_PRIV(SystemException):
    code = -39000


class SYS_NO_DATA_OBJ_PERMISSION(SystemException):
    code = -40000


class SYS_DELETE_DISALLOWED(SystemException):
    code = -41000


class SYS_OPEN_REI_FILE_ERR(SystemException):
    code = -42000


class SYS_NO_RCAT_SERVER_ERR(SystemException):
    code = -43000


class SYS_UNMATCH_PACK_INSTRUCTI_NAME(SystemException):
    code = -44000


class SYS_SVR_TO_CLI_MSI_NO_EXIST(SystemException):
    code = -45000


class SYS_COPY_ALREADY_IN_RESC(SystemException):
    code = -46000


class SYS_RECONN_OPR_MISMATCH(SystemException):
    code = -47000


class SYS_INPUT_PERM_OUT_OF_RANGE(SystemException):
    code = -48000


class SYS_FORK_ERROR(SystemException):
    code = -49000


class SYS_PIPE_ERROR(SystemException):
    code = -50000


class SYS_EXEC_CMD_STATUS_SZ_ERROR(SystemException):
    code = -51000


class SYS_PATH_IS_NOT_A_FILE(SystemException):
    code = -52000


class SYS_UNMATCHED_SPEC_COLL_TYPE(SystemException):
    code = -53000


class SYS_TOO_MANY_QUERY_RESULT(SystemException):
    code = -54000


class SYS_SPEC_COLL_NOT_IN_CACHE(SystemException):
    code = -55000


class SYS_SPEC_COLL_OBJ_NOT_EXIST(SystemException):
    code = -56000


class SYS_REG_OBJ_IN_SPEC_COLL(SystemException):
    code = -57000


class SYS_DEST_SPEC_COLL_SUB_EXIST(SystemException):
    code = -58000


class SYS_SRC_DEST_SPEC_COLL_CONFLICT(SystemException):
    code = -59000


class SYS_UNKNOWN_SPEC_COLL_CLASS(SystemException):
    code = -60000


class SYS_DUPLICATE_XMSG_TICKET(SystemException):
    code = -61000


class SYS_UNMATCHED_XMSG_TICKET(SystemException):
    code = -62000


class SYS_NO_XMSG_FOR_MSG_NUMBER(SystemException):
    code = -63000


class SYS_COLLINFO_2_FORMAT_ERR(SystemException):
    code = -64000


class SYS_CACHE_STRUCT_FILE_RESC_ERR(SystemException):
    code = -65000


class SYS_NOT_SUPPORTED(SystemException):
    code = -66000


class SYS_TAR_STRUCT_FILE_EXTRACT_ERR(SystemException):
    code = -67000


class SYS_STRUCT_FILE_DESC_ERR(SystemException):
    code = -68000


class SYS_TAR_OPEN_ERR(SystemException):
    code = -69000


class SYS_TAR_EXTRACT_ALL_ERR(SystemException):
    code = -70000


class SYS_TAR_CLOSE_ERR(SystemException):
    code = -71000


class SYS_STRUCT_FILE_PATH_ERR(SystemException):
    code = -72000


class SYS_MOUNT_MOUNTED_COLL_ERR(SystemException):
    code = -73000


class SYS_COLL_NOT_MOUNTED_ERR(SystemException):
    code = -74000


class SYS_STRUCT_FILE_BUSY_ERR(SystemException):
    code = -75000


class SYS_STRUCT_FILE_INMOUNTED_COLL(SystemException):
    code = -76000


class SYS_COPY_NOT_EXIST_IN_RESC(SystemException):
    code = -77000


class SYS_RESC_DOES_NOT_EXIST(SystemException):
    code = -78000


class SYS_COLLECTION_NOT_EMPTY(SystemException):
    code = -79000


class SYS_OBJ_TYPE_NOT_STRUCT_FILE(SystemException):
    code = -80000


class SYS_WRONG_RESC_POLICY_FOR_BUN_OPR(SystemException):
    code = -81000


class SYS_DIR_IN_VAULT_NOT_EMPTY(SystemException):
    code = -82000


class SYS_OPR_FLAG_NOT_SUPPORT(SystemException):
    code = -83000


class SYS_TAR_APPEND_ERR(SystemException):
    code = -84000


class SYS_INVALID_PROTOCOL_TYPE(SystemException):
    code = -85000


class SYS_UDP_CONNECT_ERR(SystemException):
    code = -86000


class SYS_UDP_TRANSFER_ERR(SystemException):
    code = -89000


class SYS_UDP_NO_SUPPORT_ERR(SystemException):
    code = -90000


class SYS_READ_MSG_BODY_LEN_ERR(SystemException):
    code = -91000


class CROSS_ZONE_SOCK_CONNECT_ERR(SystemException):
    code = -92000


class SYS_NO_FREE_RE_THREAD(SystemException):
    code = -93000


class SYS_BAD_RE_THREAD_INX(SystemException):
    code = -94000


class SYS_CANT_DIRECTLY_ACC_COMPOUND_RESC(SystemException):
    code = -95000


class SYS_SRC_DEST_RESC_COMPOUND_TYPE(SystemException):
    code = -96000


class SYS_CACHE_RESC_NOT_ON_SAME_HOST(SystemException):
    code = -97000


class SYS_NO_CACHE_RESC_IN_GRP(SystemException):
    code = -98000


class SYS_UNMATCHED_RESC_IN_RESC_GRP(SystemException):
    code = -99000


class SYS_CANT_MV_BUNDLE_DATA_TO_TRASH(SystemException):
    code = -100000


class SYS_CANT_MV_BUNDLE_DATA_BY_COPY(SystemException):
    code = -101000


class SYS_EXEC_TAR_ERR(SystemException):
    code = -102000


class SYS_CANT_CHKSUM_COMP_RESC_DATA(SystemException):
    code = -103000


class SYS_CANT_CHKSUM_BUNDLED_DATA(SystemException):
    code = -104000


class SYS_RESC_IS_DOWN(SystemException):
    code = -105000


class SYS_UPDATE_REPL_INFO_ERR(SystemException):
    code = -106000


class SYS_COLL_LINK_PATH_ERR(SystemException):
    code = -107000


class SYS_LINK_CNT_EXCEEDED_ERR(SystemException):
    code = -108000


class SYS_CROSS_ZONE_MV_NOT_SUPPORTED(SystemException):
    code = -109000


class SYS_RESC_QUOTA_EXCEEDED(SystemException):
    code = -110000


class SYS_INVALID_INPUT_PARAM(SystemException):
    code = -130000


class UserInputException(iRODSException):
    pass


class USER_AUTH_SCHEME_ERR(UserInputException):
    code = -300000


class USER_AUTH_STRING_EMPTY(UserInputException):
    code = -301000


class USER_RODS_HOST_EMPTY(UserInputException):
    code = -302000


class USER_RODS_HOSTNAME_ERR(UserInputException):
    code = -303000


class USER_SOCK_OPEN_ERR(UserInputException):
    code = -304000


class USER_SOCK_CONNECT_ERR(UserInputException):
    code = -305000


class USER_STRLEN_TOOLONG(UserInputException):
    code = -306000


class USER_API_INPUT_ERR(UserInputException):
    code = -307000


class USER_PACKSTRUCT_INPUT_ERR(UserInputException):
    code = -308000


class USER_NO_SUPPORT_ERR(UserInputException):
    code = -309000


class USER_FILE_DOES_NOT_EXIST(UserInputException):
    code = -310000


class USER_FILE_TOO_LARGE(UserInputException):
    code = -311000


class OVERWRITE_WITHOUT_FORCE_FLAG(UserInputException):
    code = -312000


class UNMATCHED_KEY_OR_INDEX(UserInputException):
    code = -313000


class USER_CHKSUM_MISMATCH(UserInputException):
    code = -314000


class USER_BAD_KEYWORD_ERR(UserInputException):
    code = -315000


class USER__NULL_INPUT_ERR(UserInputException):
    code = -316000


class USER_INPUT_PATH_ERR(UserInputException):
    code = -317000


class USER_INPUT_OPTION_ERR(UserInputException):
    code = -318000


class USER_INVALID_USERNAME_FORMAT(UserInputException):
    code = -319000


class USER_DIRECT_RESC_INPUT_ERR(UserInputException):
    code = -320000


class USER_NO_RESC_INPUT_ERR(UserInputException):
    code = -321000


class USER_PARAM_LABEL_ERR(UserInputException):
    code = -322000


class USER_PARAM_TYPE_ERR(UserInputException):
    code = -323000


class BASE64_BUFFER_OVERFLOW(UserInputException):
    code = -324000


class BASE64_INVALID_PACKET(UserInputException):
    code = -325000


class USER_MSG_TYPE_NO_SUPPORT(UserInputException):
    code = -326000


class USER_RSYNC_NO_MODE_INPUT_ERR(UserInputException):
    code = -337000


class USER_OPTION_INPUT_ERR(UserInputException):
    code = -338000


class SAME_SRC_DEST_PATHS_ERR(UserInputException):
    code = -339000


class USER_RESTART_FILE_INPUT_ERR(UserInputException):
    code = -340000


class RESTART_OPR_FAILED(UserInputException):
    code = -341000


class BAD_EXEC_CMD_PATH(UserInputException):
    code = -342000


class EXEC_CMD_OUTPUT_TOO_LARGE(UserInputException):
    code = -343000


class EXEC_CMD_ERROR(UserInputException):
    code = -344000


class BAD_INPUT_DESC_INDEX(UserInputException):
    code = -345000


class USER_PATH_EXCEEDS_MAX(UserInputException):
    code = -346000


class USER_SOCK_CONNECT_TIMEDOUT(UserInputException):
    code = -347000


class USER_API_VERSION_MISMATCH(UserInputException):
    code = -348000


class USER_INPUT_FORMAT_ERR(UserInputException):
    code = -349000


class USER_ACCESS_DENIED(UserInputException):
    code = -350000


class CANT_RM_MV_BUNDLE_TYPE(UserInputException):
    code = -351000


class NO_MORE_RESULT(UserInputException):
    code = -352000


class NO_KEY_WD_IN_MS_INP_STR(UserInputException):
    code = -353000


class CANT_RM_NON_EMPTY_HOME_COLL(UserInputException):
    code = -354000


class CANT_UNREG_IN_VAULT_FILE(UserInputException):
    code = -355000


class NO_LOCAL_FILE_RSYNC_IN_MSI(UserInputException):
    code = -356000


class LOCKED_DATA_OBJECT_ACCESS(SystemException):
    code = -406000


class CHECK_VERIFICATION_RESULTS(SystemException):
    code = -407000


class FileDriverException(iRODSException):
    pass


class FILE_INDEX_LOOKUP_ERR(FileDriverException):
    code = -500000


class UNIX_FILE_OPEN_ERR(FileDriverException):
    code = -510000


class UNIX_FILE_CREATE_ERR(FileDriverException):
    code = -511000


class UNIX_FILE_READ_ERR(FileDriverException):
    code = -512000


class UNIX_FILE_WRITE_ERR(FileDriverException):
    code = -513000


class UNIX_FILE_CLOSE_ERR(FileDriverException):
    code = -514000


class UNIX_FILE_UNLINK_ERR(FileDriverException):
    code = -515000


class UNIX_FILE_STAT_ERR(FileDriverException):
    code = -516000


class UNIX_FILE_FSTAT_ERR(FileDriverException):
    code = -517000


class UNIX_FILE_LSEEK_ERR(FileDriverException):
    code = -518000


class UNIX_FILE_FSYNC_ERR(FileDriverException):
    code = -519000


class UNIX_FILE_MKDIR_ERR(FileDriverException):
    code = -520000


class UNIX_FILE_RMDIR_ERR(FileDriverException):
    code = -521000


class UNIX_FILE_OPENDIR_ERR(FileDriverException):
    code = -522000


class UNIX_FILE_CLOSEDIR_ERR(FileDriverException):
    code = -523000


class UNIX_FILE_READDIR_ERR(FileDriverException):
    code = -524000


class UNIX_FILE_STAGE_ERR(FileDriverException):
    code = -525000


class UNIX_FILE_GET_FS_FREESPACE_ERR(FileDriverException):
    code = -526000


class UNIX_FILE_CHMOD_ERR(FileDriverException):
    code = -527000


class UNIX_FILE_RENAME_ERR(FileDriverException):
    code = -528000


class UNIX_FILE_TRUNCATE_ERR(FileDriverException):
    code = -529000


class UNIX_FILE_LINK_ERR(FileDriverException):
    code = -530000


class UniversalMSSDriverException(FileDriverException):
    pass


class UNIV_MSS_SYNCTOARCH_ERR(UniversalMSSDriverException):
    code = -550000


class UNIV_MSS_STAGETOCACHE_ERR(UniversalMSSDriverException):
    code = -551000


class UNIV_MSS_UNLINK_ERR(UniversalMSSDriverException):
    code = -552000


class UNIV_MSS_MKDIR_ERR(UniversalMSSDriverException):
    code = -553000


class UNIV_MSS_CHMOD_ERR(UniversalMSSDriverException):
    code = -554000


class UNIV_MSS_STAT_ERR(UniversalMSSDriverException):
    code = -555000


class HPSSDriverException(FileDriverException):
    pass


class HPSS_AUTH_NOT_SUPPORTED(HPSSDriverException):
    code = -600000


class HPSS_FILE_OPEN_ERR(HPSSDriverException):
    code = -610000


class HPSS_FILE_CREATE_ERR(HPSSDriverException):
    code = -611000


class HPSS_FILE_READ_ERR(HPSSDriverException):
    code = -612000


class HPSS_FILE_WRITE_ERR(HPSSDriverException):
    code = -613000


class HPSS_FILE_CLOSE_ERR(HPSSDriverException):
    code = -614000


class HPSS_FILE_UNLINK_ERR(HPSSDriverException):
    code = -615000


class HPSS_FILE_STAT_ERR(HPSSDriverException):
    code = -616000


class HPSS_FILE_FSTAT_ERR(HPSSDriverException):
    code = -617000


class HPSS_FILE_LSEEK_ERR(HPSSDriverException):
    code = -618000


class HPSS_FILE_FSYNC_ERR(HPSSDriverException):
    code = -619000


class HPSS_FILE_MKDIR_ERR(HPSSDriverException):
    code = -620000


class HPSS_FILE_RMDIR_ERR(HPSSDriverException):
    code = -621000


class HPSS_FILE_OPENDIR_ERR(HPSSDriverException):
    code = -622000


class HPSS_FILE_CLOSEDIR_ERR(HPSSDriverException):
    code = -623000


class HPSS_FILE_READDIR_ERR(HPSSDriverException):
    code = -624000


class HPSS_FILE_STAGE_ERR(HPSSDriverException):
    code = -625000


class HPSS_FILE_GET_FS_FREESPACE_ERR(HPSSDriverException):
    code = -626000


class HPSS_FILE_CHMOD_ERR(HPSSDriverException):
    code = -627000


class HPSS_FILE_RENAME_ERR(HPSSDriverException):
    code = -628000


class HPSS_FILE_TRUNCATE_ERR(HPSSDriverException):
    code = -629000


class HPSS_FILE_LINK_ERR(HPSSDriverException):
    code = -630000


class HPSS_AUTH_ERR(HPSSDriverException):
    code = -631000


class HPSS_WRITE_LIST_ERR(HPSSDriverException):
    code = -632000


class HPSS_READ_LIST_ERR(HPSSDriverException):
    code = -633000


class HPSS_TRANSFER_ERR(HPSSDriverException):
    code = -634000


class HPSS_MOVER_PROT_ERR(HPSSDriverException):
    code = -635000


class AmazonS3Exception(FileDriverException):
    pass


class S3_INIT_ERROR(AmazonS3Exception):
    code = -701000


class S3_PUT_ERROR(AmazonS3Exception):
    code = -702000


class S3_GET_ERROR(AmazonS3Exception):
    code = -703000


class S3_FILE_UNLINK_ERR(AmazonS3Exception):
    code = -715000


class S3_FILE_STAT_ERR(AmazonS3Exception):
    code = -716000


class S3_FILE_COPY_ERR(AmazonS3Exception):
    code = -717000


class CatalogLibraryException(iRODSException):
    pass


class CATALOG_NOT_CONNECTED(CatalogLibraryException):
    code = -801000


class CAT_ENV_ERR(CatalogLibraryException):
    code = -802000


class CAT_CONNECT_ERR(CatalogLibraryException):
    code = -803000


class CAT_DISCONNECT_ERR(CatalogLibraryException):
    code = -804000


class CAT_CLOSE_ENV_ERR(CatalogLibraryException):
    code = -805000


class CAT_SQL_ERR(CatalogLibraryException):
    code = -806000


class CAT_GET_ROW_ERR(CatalogLibraryException):
    code = -807000


class CAT_NO_ROWS_FOUND(CatalogLibraryException):
    code = -808000


class CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME(CatalogLibraryException):
    code = -809000

class CAT_NO_CHECKSUM_FOR_REPLICA (CatalogLibraryException):
    code = -862000

class CAT_INVALID_RESOURCE_TYPE(CatalogLibraryException):
    code = -810000


class CAT_INVALID_RESOURCE_CLASS(CatalogLibraryException):
    code = -811000


class CAT_INVALID_RESOURCE_NET_ADDR(CatalogLibraryException):
    code = -812000


class CAT_INVALID_RESOURCE_VAULT_PATH(CatalogLibraryException):
    code = -813000


class CAT_UNKNOWN_COLLECTION(CatalogLibraryException):
    code = -814000


class CAT_INVALID_DATA_TYPE(CatalogLibraryException):
    code = -815000


class CAT_INVALID_ARGUMENT(CatalogLibraryException):
    code = -816000


class CAT_UNKNOWN_FILE(CatalogLibraryException):
    code = -817000


class CAT_NO_ACCESS_PERMISSION(CatalogLibraryException):
    code = -818000


class CAT_SUCCESS_BUT_WITH_NO_INFO(CatalogLibraryException):
    code = -819000


class CAT_INVALID_USER_TYPE(CatalogLibraryException):
    code = -820000


class CAT_COLLECTION_NOT_EMPTY(CatalogLibraryException):
    code = -821000


class CAT_TOO_MANY_TABLES(CatalogLibraryException):
    code = -822000


class CAT_UNKNOWN_TABLE(CatalogLibraryException):
    code = -823000


class CAT_NOT_OPEN(CatalogLibraryException):
    code = -824000


class CAT_FAILED_TO_LINK_TABLES(CatalogLibraryException):
    code = -825000


class CAT_INVALID_AUTHENTICATION(CatalogLibraryException):
    code = -826000


class CAT_INVALID_USER(CatalogLibraryException):
    code = -827000


class CAT_INVALID_ZONE(CatalogLibraryException):
    code = -828000


class CAT_INVALID_GROUP(CatalogLibraryException):
    code = -829000


class CAT_INSUFFICIENT_PRIVILEGE_LEVEL(CatalogLibraryException):
    code = -830000


class CAT_INVALID_RESOURCE(CatalogLibraryException):
    code = -831000


class CAT_INVALID_CLIENT_USER(CatalogLibraryException):
    code = -832000


class CAT_NAME_EXISTS_AS_COLLECTION(CatalogLibraryException):
    code = -833000


class CAT_NAME_EXISTS_AS_DATAOBJ(CatalogLibraryException):
    code = -834000


class CAT_RESOURCE_NOT_EMPTY(CatalogLibraryException):
    code = -835000


class CAT_NOT_A_DATAOBJ_AND_NOT_A_COLLECTION(CatalogLibraryException):
    code = -836000


class CAT_RECURSIVE_MOVE(CatalogLibraryException):
    code = -837000


class CAT_LAST_REPLICA(CatalogLibraryException):
    code = -838000


class CAT_OCI_ERROR(CatalogLibraryException):
    code = -839000


class CAT_PASSWORD_EXPIRED(CatalogLibraryException):
    code = -840000


class CAT_PASSWORD_ENCODING_ERROR(CatalogLibraryException):
    code = -850000


class CAT_TABLE_ACCESS_DENIED(CatalogLibraryException):
    code = -851000


class CAT_UNKNOWN_SPECIFIC_QUERY(CatalogLibraryException):
    code = -853000

class CAT_STATEMENT_TABLE_FULL(CatalogLibraryException):
    code = -860000


class RDSException(iRODSException):
    pass


class RDA_NOT_COMPILED_IN(RDSException):
    code = -880000


class RDA_NOT_CONNECTED(RDSException):
    code = -881000


class RDA_ENV_ERR(RDSException):
    code = -882000


class RDA_CONNECT_ERR(RDSException):
    code = -883000


class RDA_DISCONNECT_ERR(RDSException):
    code = -884000


class RDA_CLOSE_ENV_ERR(RDSException):
    code = -885000


class RDA_SQL_ERR(RDSException):
    code = -886000


class RDA_CONFIG_FILE_ERR(RDSException):
    code = -887000


class RDA_ACCESS_PROHIBITED(RDSException):
    code = -888000


class RDA_NAME_NOT_FOUND(RDSException):
    code = -889000


class TicketException(CatalogLibraryException):
    pass


class CAT_TICKET_INVALID(TicketException):
    code = -890000


class CAT_TICKET_EXPIRED(TicketException):
    code = -891000


class MiscException(iRODSException):
    pass


class FILE_OPEN_ERR(MiscException):
    code = -900000


class FILE_READ_ERR(MiscException):
    code = -901000


class FILE_WRITE_ERR(MiscException):
    code = -902000


class PASSWORD_EXCEEDS_MAX_SIZE(MiscException):
    code = -903000


class ENVIRONMENT_VAR_HOME_NOT_DEFINED(MiscException):
    code = -904000


class UNABLE_TO_STAT_FILE(MiscException):
    code = -905000


class AUTH_FILE_NOT_ENCRYPTED(MiscException):
    code = -906000


class AUTH_FILE_DOES_NOT_EXIST(MiscException):
    code = -907000


class UNLINK_FAILED(MiscException):
    code = -908000


class NO_PASSWORD_ENTERED(MiscException):
    code = -909000


class REMOTE_SERVER_AUTHENTICATION_FAILURE(MiscException):
    code = -910000


class REMOTE_SERVER_AUTH_NOT_PROVIDED(MiscException):
    code = -911000


class REMOTE_SERVER_AUTH_EMPTY(MiscException):
    code = -912000


class REMOTE_SERVER_SID_NOT_DEFINED(MiscException):
    code = -913000


class GSIKRBException(iRODSException):
    pass


class GSIException(GSIKRBException):
    pass


class GSI_NOT_COMPILED_IN(GSIException):
    code = -921000


class GSI_NOT_BUILT_INTO_CLIENT(GSIException):
    code = -922000


class GSI_NOT_BUILT_INTO_SERVER(GSIException):
    code = -923000


class GSI_ERROR_IMPORT_NAME(GSIException):
    code = -924000


class GSI_ERROR_INIT_SECURITY_CONTEXT(GSIException):
    code = -925000


class GSI_ERROR_SENDING_TOKEN_LENGTH(GSIException):
    code = -926000


class GSI_ERROR_READING_TOKEN_LENGTH(GSIException):
    code = -927000


class GSI_ERROR_TOKEN_TOO_LARGE(GSIException):
    code = -928000


class GSI_ERROR_BAD_TOKEN_RCVED(GSIException):
    code = -929000


class GSI_SOCKET_READ_ERROR(GSIException):
    code = -930000


class GSI_PARTIAL_TOKEN_READ(GSIException):
    code = -931000


class GSI_SOCKET_WRITE_ERROR(GSIException):
    code = -932000


class GSI_ERROR_FROM_GSI_LIBRARY(GSIException):
    code = -933000


class GSI_ERROR_IMPORTING_NAME(GSIException):
    code = -934000


class GSI_ERROR_ACQUIRING_CREDS(GSIException):
    code = -935000


class GSI_ACCEPT_SEC_CONTEXT_ERROR(GSIException):
    code = -936000


class GSI_ERROR_DISPLAYING_NAME(GSIException):
    code = -937000


class GSI_ERROR_RELEASING_NAME(GSIException):
    code = -938000


class GSI_DN_DOES_NOT_MATCH_USER(GSIException):
    code = -939000


class GSI_QUERY_INTERNAL_ERROR(GSIException):
    code = -940000


class GSI_NO_MATCHING_DN_FOUND(GSIException):
    code = -941000


class GSI_MULTIPLE_MATCHING_DN_FOUND(GSIException):
    code = -942000


class KRBException(GSIKRBException):
    pass


class KRB_NOT_COMPILED_IN(KRBException):
    code = -951000


class KRB_NOT_BUILT_INTO_CLIENT(KRBException):
    code = -952000


class KRB_NOT_BUILT_INTO_SERVER(KRBException):
    code = -953000


class KRB_ERROR_IMPORT_NAME(KRBException):
    code = -954000


class KRB_ERROR_INIT_SECURITY_CONTEXT(KRBException):
    code = -955000


class KRB_ERROR_SENDING_TOKEN_LENGTH(KRBException):
    code = -956000


class KRB_ERROR_READING_TOKEN_LENGTH(KRBException):
    code = -957000


class KRB_ERROR_TOKEN_TOO_LARGE(KRBException):
    code = -958000


class KRB_ERROR_BAD_TOKEN_RCVED(KRBException):
    code = -959000


class KRB_SOCKET_READ_ERROR(KRBException):
    code = -960000


class KRB_PARTIAL_TOKEN_READ(KRBException):
    code = -961000


class KRB_SOCKET_WRITE_ERROR(KRBException):
    code = -962000


class KRB_ERROR_FROM_KRB_LIBRARY(KRBException):
    code = -963000


class KRB_ERROR_IMPORTING_NAME(KRBException):
    code = -964000


class KRB_ERROR_ACQUIRING_CREDS(KRBException):
    code = -965000


class KRB_ACCEPT_SEC_CONTEXT_ERROR(KRBException):
    code = -966000


class KRB_ERROR_DISPLAYING_NAME(KRBException):
    code = -967000


class KRB_ERROR_RELEASING_NAME(KRBException):
    code = -968000


class KRB_USER_DN_NOT_FOUND(KRBException):
    code = -969000


class KRB_NAME_MATCHES_MULTIPLE_USERS(KRBException):
    code = -970000


class KRB_QUERY_INTERNAL_ERROR(KRBException):
    code = -971000


class RuleEngineException(iRODSException):
    pass


class OBJPATH_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1000000


class RESCNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1001000


class DATATYPE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1002000


class DATASIZE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1003000


class CHKSUM_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1004000


class VERSION_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1005000


class FILEPATH_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1006000


class REPLNUM_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1007000


class REPLSTATUS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1008000


class DATAOWNER_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1009000


class DATAOWNERZONE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1010000


class DATAEXPIRY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1011000


class DATACOMMENTS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1012000


class DATACREATE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1013000


class DATAMODIFY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1014000


class DATAACCESS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1015000


class DATAACCESSINX_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1016000


class RE_TYPE_ERROR(RuleEngineException):
    code = -1230000


class NO_RULE_FOUND_ERR(RuleEngineException):
    code = -1017000


class NO_MORE_RULES_ERR(RuleEngineException):
    code = -1018000


class UNMATCHED_ACTION_ERR(RuleEngineException):
    code = -1019000


class RULES_FILE_READ_ERROR(RuleEngineException):
    code = -1020000


class ACTION_ARG_COUNT_MISMATCH(RuleEngineException):
    code = -1021000


class MAX_NUM_OF_ARGS_IN_ACTION_EXCEEDED(RuleEngineException):
    code = -1022000


class UNKNOWN_PARAM_IN_RULE_ERR(RuleEngineException):
    code = -1023000


class DESTRESCNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1024000


class BACKUPRESCNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1025000


class DATAID_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1026000


class COLLID_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1027000


class RESCGROUPNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1028000


class STATUSSTRING_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1029000


class DATAMAPID_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1030000


class USERNAMECLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1031000


class RODSZONECLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1032000


class USERTYPECLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1033000


class HOSTCLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1034000


class AUTHSTRCLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1035000


class USERAUTHSCHEMECLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1036000


class USERINFOCLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1037000


class USERCOMMENTCLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1038000


class USERCREATECLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1039000


class USERMODIFYCLIENT_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1040000


class USERNAMEPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1041000


class RODSZONEPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1042000


class USERTYPEPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1043000


class HOSTPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1044000


class AUTHSTRPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1045000


class USERAUTHSCHEMEPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1046000


class USERINFOPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1047000


class USERCOMMENTPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1048000


class USERCREATEPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1049000


class USERMODIFYPROXY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1050000


class COLLNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1051000


class COLLPARENTNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1052000


class COLLOWNERNAME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1053000


class COLLOWNERZONE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1054000


class COLLEXPIRY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1055000


class COLLCOMMENTS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1056000


class COLLCREATE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1057000


class COLLMODIFY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1058000


class COLLACCESS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1059000


class COLLACCESSINX_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1060000


class COLLMAPID_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1062000


class COLLINHERITANCE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1063000


class RESCZONE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1065000


class RESCLOC_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1066000


class RESCTYPE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1067000


class RESCTYPEINX_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1068000


class RESCCLASS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1069000


class RESCCLASSINX_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1070000


class RESCVAULTPATH_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1071000


class NUMOPEN_ORTS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1072000


class PARAOPR_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1073000


class RESCID_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1074000


class GATEWAYADDR_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1075000


class RESCMAX_BJSIZE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1076000


class FREESPACE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1077000


class FREESPACETIME_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1078000


class FREESPACETIMESTAMP_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1079000


class RESCINFO_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1080000


class RESCCOMMENTS_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1081000


class RESCCREATE_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1082000


class RESCMODIFY_EMPTY_IN_STRUCT_ERR(RuleEngineException):
    code = -1083000


class INPUT_ARG_NOT_WELL_FORMED_ERR(RuleEngineException):
    code = -1084000


class INPUT_ARG_OUT_OF_ARGC_RANGE_ERR(RuleEngineException):
    code = -1085000


class INSUFFICIENT_INPUT_ARG_ERR(RuleEngineException):
    code = -1086000


class INPUT_ARG_DOES_NOT_MATCH_ERR(RuleEngineException):
    code = -1087000


class RETRY_WITHOUT_RECOVERY_ERR(RuleEngineException):
    code = -1088000


class CUT_ACTION_PROCESSED_ERR(RuleEngineException):
    code = -1089000


class ACTION_FAILED_ERR(RuleEngineException):
    code = -1090000


class FAIL_ACTION_ENCOUNTERED_ERR(RuleEngineException):
    code = -1091000


class VARIABLE_NAME_TOO_LONG_ERR(RuleEngineException):
    code = -1092000


class UNKNOWN_VARIABLE_MAP_ERR(RuleEngineException):
    code = -1093000


class UNDEFINED_VARIABLE_MAP_ERR(RuleEngineException):
    code = -1094000


class NULL_VALUE_ERR(RuleEngineException):
    code = -1095000


class DVARMAP_FILE_READ_ERROR(RuleEngineException):
    code = -1096000


class NO_RULE_OR_MSI_FUNCTION_FOUND_ERR(RuleEngineException):
    code = -1097000


class FILE_CREATE_ERROR(RuleEngineException):
    code = -1098000


class FMAP_FILE_READ_ERROR(RuleEngineException):
    code = -1099000


class DATE_FORMAT_ERR(RuleEngineException):
    code = -1100000


class RULE_FAILED_ERR(RuleEngineException):
    code = -1101000


class NO_MICROSERVICE_FOUND_ERR(RuleEngineException):
    code = -1102000


class INVALID_REGEXP(RuleEngineException):
    code = -1103000


class INVALID_OBJECT_NAME(RuleEngineException):
    code = -1104000


class INVALID_OBJECT_TYPE(RuleEngineException):
    code = -1105000


class NO_VALUES_FOUND(RuleEngineException):
    code = -1106000


class NO_COLUMN_NAME_FOUND(RuleEngineException):
    code = -1107000


class BREAK_ACTION_ENCOUNTERED_ERR(RuleEngineException):
    code = -1108000


class CUT_ACTION_ON_SUCCESS_PROCESSED_ERR(RuleEngineException):
    code = -1109000


class MSI_OPERATION_NOT_ALLOWED(RuleEngineException):
    code = -1110000


class RULE_ENGINE_ERROR(RuleEngineException):
    code = -1828000


class PHPException(iRODSException):
    pass


class PHP_EXEC_SCRIPT_ERR(PHPException):
    code = -1600000


class PHP_REQUEST_STARTUP_ERR(PHPException):
    code = -1601000


class PHP_OPEN_SCRIPT_FILE_ERR(PHPException):
    code = -1602000

class PAMException(iRODSException):
    pass


class PAM_AUTH_NOT_BUILT_INTO_CLIENT(PAMException):
    code = -991000


class PAM_AUTH_NOT_BUILT_INTO_SERVER(PAMException):
    code = -992000


class PAM_AUTH_PASSWORD_FAILED(PAMException):
    code = -993000


class PAM_AUTH_PASSWORD_INVALID_TTL(PAMException):
    code = -994000



from __future__ import print_function
from __future__ import absolute_import
import logging
from os.path import dirname, basename

from irods.manager import Manager
from irods.message import MetadataRequest, iRODSMessage, JSON_Message
from irods.api_number import api_number
from irods.models import (DataObject, Collection, Resource,
                          User, DataObjectMeta, CollectionMeta, ResourceMeta, UserMeta)
from irods.meta import iRODSMeta, AVUOperation


logger = logging.getLogger(__name__)


class InvalidAtomicAVURequest(Exception): pass


class MetadataManager(Manager):

    @staticmethod
    def _model_class_to_resource_type(model_cls):
        return {
            DataObject: 'd',
            Collection: 'c',
            Resource: 'r',
            User: 'u',
        }[model_cls]

    @staticmethod
    def _model_class_to_resource_description(model_cls):
        return {
            DataObject: 'data_object',
            Collection: 'collection',
            Resource: 'resource',
            User: 'user',
        }[model_cls]

    def get(self, model_cls, path):
        resource_type = self._model_class_to_resource_type(model_cls)
        model = {
            'd': DataObjectMeta,
            'c': CollectionMeta,
            'r': ResourceMeta,
            'u': UserMeta
        }[resource_type]
        conditions = {
            'd': [
                Collection.name == dirname(path),
                DataObject.name == basename(path)
            ],
            'c': [Collection.name == path],
            'r': [Resource.name == path],
            'u': [User.name == path]
        }[resource_type]
        results = self.sess.query(model.id, model.name, model.value, model.units)\
            .filter(*conditions).all()
        return [iRODSMeta(
            row[model.name],
            row[model.value],
            row[model.units],
            avu_id=row[model.id]
        ) for row in results]

    def add(self, model_cls, path, meta):
        # Avoid sending request with empty argument(s)
        if not(len(path) and len(meta.name) and len(meta.value)):
            raise ValueError('Empty value in ' + repr(meta))

        resource_type = self._model_class_to_resource_type(model_cls)
        message_body = MetadataRequest(
            "add",
            "-" + resource_type,
            path,
            meta.name,
            meta.value,
            meta.units
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['MOD_AVU_METADATA_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    def remove(self, model_cls, path, meta):
        resource_type = self._model_class_to_resource_type(model_cls)
        message_body = MetadataRequest(
            "rm",
            "-" + resource_type,
            path,
            meta.name,
            meta.value,
            meta.units
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['MOD_AVU_METADATA_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    def copy(self, src_model_cls, dest_model_cls, src, dest):
        src_resource_type = self._model_class_to_resource_type(src_model_cls)
        dest_resource_type = self._model_class_to_resource_type(dest_model_cls)
        message_body = MetadataRequest(
            "cp",
            "-" + src_resource_type,
            "-" + dest_resource_type,
            src,
            dest
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['MOD_AVU_METADATA_AN'])

        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    def set(self, model_cls, path, meta):
        resource_type = self._model_class_to_resource_type(model_cls)
        message_body = MetadataRequest(
            "set",
            "-" + resource_type,
            path,
            meta.name,
            meta.value,
            meta.units
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['MOD_AVU_METADATA_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    @staticmethod
    def _avu_operation_to_dict( op ):
        opJSON = { "operation": op.operation,
                   "attribute": op.avu.name,
                   "value": op.avu.value
        }
        if op.avu.units not in ("",None):
            opJSON["units"] = op.avu.units
        return opJSON

    def apply_atomic_operations(self, model_cls, path, *avu_ops):
        if not all(isinstance(op,AVUOperation) for op in avu_ops):
            raise InvalidAtomicAVURequest("avu_ops must contain 1 or more AVUOperations")
        request = {
            "entity_name": path,
            "entity_type": self._model_class_to_resource_description(model_cls),
            "operations" : [self._avu_operation_to_dict(op) for op in avu_ops]
        }
        self._call_atomic_metadata_api(request)

    def _call_atomic_metadata_api(self, request_text):
        with self.sess.pool.get_connection() as conn:
            request_msg = iRODSMessage("RODS_API_REQ",  JSON_Message( request_text, conn.server_version ),
                                       int_info=api_number['ATOMIC_APPLY_METADATA_OPERATIONS_APN'])
            conn.send( request_msg )
            response = conn.recv()
        response_msg = response.get_json_encoded_struct()
        logger.debug("in atomic_metadata, server responded with: %r",response_msg)




#! /usr/bin/env python
from __future__ import print_function
from __future__ import absolute_import
import os
import sys
import unittest
import textwrap
import json
import shutil
import ssl
import irods.test.helpers as helpers
from irods.connection import Connection
from irods.session import iRODSSession
from irods.rule import Rule
from irods.models import User
from socket import gethostname
from irods.password_obfuscation import (encode as pw_encode)
from irods.connection import PlainTextPAMPasswordError
from irods.access import iRODSAccess
import contextlib
import socket
from re import compile as regex
try:
    from re import _pattern_type as regex_type
except ImportError:
    from re import Pattern as regex_type  # Python 3.7+


def json_file_update(fname,keys_to_delete=(),**kw):
    with open(fname,'r') as f:
        j = json.load(f)
    j.update(**kw)
    for k in keys_to_delete:
        if k in j: del j [k]
        elif isinstance(k,regex_type):
            jk = [i for i in j.keys() if k.search(i)]
            for ky in jk: del j[ky]
    with open(fname,'w') as out:
        json.dump(j, out, indent=4)

def env_dir_fullpath(authtype):  return os.path.join( os.environ['HOME'] , '.irods.' + authtype)
def json_env_fullpath(authtype):  return os.path.join( env_dir_fullpath(authtype), 'irods_environment.json')
def secrets_fullpath(authtype):   return os.path.join( env_dir_fullpath(authtype), '.irodsA')

SERVER_ENV_PATH = os.path.expanduser('~irods/.irods/irods_environment.json')

SERVER_ENV_SSL_SETTINGS = {
    "irods_ssl_certificate_chain_file": "/etc/irods/ssl/irods.crt",
    "irods_ssl_certificate_key_file": "/etc/irods/ssl/irods.key",
    "irods_ssl_dh_params_file": "/etc/irods/ssl/dhparams.pem",
    "irods_ssl_ca_certificate_file": "/etc/irods/ssl/irods.crt",
    "irods_ssl_verify_server": "cert"
}

def update_service_account_for_SSL():
    json_file_update( SERVER_ENV_PATH, **SERVER_ENV_SSL_SETTINGS )

CLIENT_OPTIONS_FOR_SSL = {
    "irods_client_server_policy": "CS_NEG_REQUIRE",
    "irods_client_server_negotiation": "request_server_negotiation",
    "irods_ssl_ca_certificate_file": "/etc/irods/ssl/irods.crt",
    "irods_ssl_verify_server": "cert",
    "irods_encryption_key_size": 16,
    "irods_encryption_salt_size": 8,
    "irods_encryption_num_hash_rounds": 16,
    "irods_encryption_algorithm": "AES-256-CBC"
}


def client_env_from_server_env(user_name, auth_scheme=""):
    cli_env = {}
    with open(SERVER_ENV_PATH) as f:
        srv_env = json.load(f)
        for k in [ "irods_host", "irods_zone_name", "irods_port"  ]:
            cli_env [k] = srv_env[k]
    cli_env["irods_user_name"] = user_name
    if auth_scheme:
        cli_env["irods_authentication_scheme"] = auth_scheme
    return cli_env

@contextlib.contextmanager
def pam_password_in_plaintext(allow=True):
    saved = bool(Connection.DISALLOWING_PAM_PLAINTEXT)
    try:
        Connection.DISALLOWING_PAM_PLAINTEXT = not(allow)
        yield
    finally:
        Connection.DISALLOWING_PAM_PLAINTEXT = saved


class TestLogins(unittest.TestCase):
    '''
    Ideally, these tests should move into CI, but that would require the server
    (currently a different node than the client) to have SSL certs created and
    enabled.

    Until then, we require these tests to be run manually on a server node,
    with:

        python -m unittest "irods.test.login_auth_test[.XX[.YY]]'

    Additionally:

      1. The PAM/SSL tests under the TestLogins class should be run on a
         single-node iRODS system, by the service account user. This ensures
         the /etc/irods directory is local and writable.

      2. ./setupssl.py (sets up SSL keys etc. in /etc/irods/ssl) should be run
         first to create (or overwrite, if appropriate) the /etc/irods/ssl directory
         and its contents.

      3. Must add & override configuration entries in /var/lib/irods/irods_environment
         Per https://slides.com/irods/ugm2018-ssl-and-pam-configuration#/3/7

      4. Create rodsuser alissa and corresponding unix user with the appropriate
         passwords as below.
    '''

    test_rods_user = 'alissa'

    user_auth_envs = {
        '.irods.pam': {
            'USER':     test_rods_user,
            'PASSWORD': 'test123', # UNIX pw
            'AUTH':     'pam'
        },
        '.irods.native': {
            'USER':     test_rods_user,
            'PASSWORD': 'apass',   # iRODS pw
            'AUTH':     'native'
        }
    }

    env_save = {}

    @contextlib.contextmanager
    def setenv(self,var,newvalue):
        try:
            self.env_save[var] = os.environ.get(var,None)
            os.environ[var] = newvalue
            yield newvalue
        finally:
            oldvalue = self.env_save[var]
            if oldvalue is None:
                del os.environ[var]
            else:
                os.environ[var]=oldvalue

    @classmethod
    def create_env_dirs(cls):
        dirs = {}
        retval = []
        # -- create environment configurations and secrets
        with pam_password_in_plaintext():
            for dirname,lookup in cls.user_auth_envs.items():
                if lookup['AUTH'] == 'pam':
                    ses = iRODSSession( host=gethostname(),
                                        user=lookup['USER'],
                                        zone='tempZone',
                                        authentication_scheme=lookup['AUTH'],
                                        password=lookup['PASSWORD'],
                                        port= 1247 )
                    try:
                        pam_hashes = ses.pam_pw_negotiated
                    except AttributeError:
                        pam_hashes = []
                    if not pam_hashes: print('Warning ** PAM pw couldnt be generated' ); break
                    scrambled_pw = pw_encode( pam_hashes[0] )
               #elif lookup['AUTH'] == 'XXXXXX': # TODO: insert other authentication schemes here
                elif lookup['AUTH'] in ('native', '',None):
                    scrambled_pw = pw_encode( lookup['PASSWORD'] )
                cl_env = client_env_from_server_env(cls.test_rods_user)
                if lookup.get('AUTH',None) is not None:     # - specify auth scheme only if given
                    cl_env['irods_authentication_scheme'] = lookup['AUTH']
                dirbase = os.path.join(os.environ['HOME'],dirname)
                dirs[dirbase] = { 'secrets':scrambled_pw , 'client_environment':cl_env }

        # -- create the environment directories and write into them the configurations just created
        for absdir in dirs.keys():
            shutil.rmtree(absdir,ignore_errors=True)
            os.mkdir(absdir)
            with open(os.path.join(absdir,'irods_environment.json'),'w') as envfile:
                envfile.write('{}')
            json_file_update(envfile.name, **dirs[absdir]['client_environment'])
            with open(os.path.join(absdir,'.irodsA'),'w') as secrets_file:
                secrets_file.write(dirs[absdir]['secrets'])
            os.chmod(secrets_file.name,0o600)

        retval = dirs.keys()
        return retval


    @staticmethod
    def get_server_ssl_negotiation( session ):

        rule_body = textwrap.dedent('''
                                    test { *out=""; acPreConnect(*out);
                                               writeLine("stdout", "*out");
                                         }
                                    ''')
        myrule = Rule(session, body=rule_body, params={}, output='ruleExecOut')
        out_array = myrule.execute()
        buf = out_array.MsParam_PI[0].inOutStruct.stdoutBuf.buf.decode('utf-8')
        eol_offset = buf.find('\n')
        return  buf[:eol_offset]  if  eol_offset >= 0  else None

    @classmethod
    def setUpClass(cls):
        cls.admin = helpers.make_session()
        if cls.test_rods_user in (row[User.name] for row in cls.admin.query(User.name)):
            cls.server_ssl_setting = cls.get_server_ssl_negotiation( cls.admin )
            cls.envdirs = cls.create_env_dirs()
            if not cls.envdirs:
                raise RuntimeError('Could not create one or more client environments')

    @classmethod
    def tearDownClass(cls):
        for envdir in getattr(cls, 'envdirs', []):
            shutil.rmtree(envdir, ignore_errors=True)
        cls.admin.cleanup()

    def setUp(self):
        if not getattr(self, 'envdirs', []):
            self.skipTest('The test_rods_user "{}" does not exist'.format(self.test_rods_user))
        if os.environ['HOME'] != '/var/lib/irods':
            self.skipTest('Must be run as irods')
        super(TestLogins,self).setUp()

    def tearDown(self):
        super(TestLogins,self).tearDown()

    def validate_session(self, session, verbose=False, **options):
        
        # - try to get the home collection
        home_coll =  '/{0.zone}/home/{0.username}'.format(session)
        self.assertTrue(session.collections.get(home_coll).path == home_coll)
        if verbose: print(home_coll)
        # - check user is as expected
        self.assertEqual( session.username, self.test_rods_user )
        # - check socket type (normal vs SSL) against whether ssl requested
        use_ssl = options.pop('ssl',None)
        if use_ssl is not None:
            my_connect = [s for s in (session.pool.active|session.pool.idle)] [0]
            self.assertEqual( bool( use_ssl ), my_connect.socket.__class__ is ssl.SSLSocket )


#   def test_demo(self): self.demo()

#   def demo(self): # for future reference - skipping based on CS_NEG_DONT_CARE setting
#       if self.server_ssl_setting == 'CS_NEG_DONT_CARE':
#           self.skipTest('skipping  b/c setting is DONT_CARE')
#       self.assertTrue (False)


    def tst0(self, ssl_opt, auth_opt, env_opt ):
        auth_opt_explicit = 'native' if auth_opt=='' else  auth_opt
        verbosity=False
        #verbosity='' # -- debug - sanity check by printing out options applied
        out = {'':''}
        if env_opt:
            with self.setenv('IRODS_ENVIRONMENT_FILE', json_env_fullpath(auth_opt_explicit)) as env_file,\
                 self.setenv('IRODS_AUTHENTICATION_FILE', secrets_fullpath(auth_opt_explicit)):
                cli_env_extras = {} if not(ssl_opt) else dict( CLIENT_OPTIONS_FOR_SSL )
                if auth_opt:
                    cli_env_extras.update( irods_authentication_scheme = auth_opt )
                    remove=[]
                else:
                    remove=[regex('authentication_')]
                with helpers.file_backed_up(env_file):
                    json_file_update( env_file, keys_to_delete=remove, **cli_env_extras )
                    session = iRODSSession(irods_env_file=env_file)
                    with open(env_file) as f:
                        out =  json.load(f)
                    self.validate_session( session, verbose = verbosity, ssl = ssl_opt )
                    session.cleanup()
            out['ARGS']='no'
        else:
            session_options = {}
            if auth_opt:
                session_options.update (authentication_scheme = auth_opt)
            if ssl_opt:
                SSL_cert = CLIENT_OPTIONS_FOR_SSL["irods_ssl_ca_certificate_file"]
                session_options.update(
                    ssl_context = ssl.create_default_context ( purpose = ssl.Purpose.SERVER_AUTH,
                                                               capath = None,
                                                               cadata = None,
                                                               cafile = SSL_cert),
                    **CLIENT_OPTIONS_FOR_SSL )
            lookup = self.user_auth_envs ['.irods.'+('native' if not(auth_opt) else auth_opt)]
            session = iRODSSession ( host=gethostname(),
                                     user=lookup['USER'],
                                     zone='tempZone',
                                     password=lookup['PASSWORD'],
                                     port= 1247,
                                     **session_options )
            out = session_options
            self.validate_session( session, verbose = verbosity, ssl = ssl_opt )
            session.cleanup()
            out['ARGS']='yes'

        if verbosity == '':
            print ('--- ssl:',ssl_opt,'/ auth:',repr(auth_opt),'/ env:',env_opt)
            print ('--- > ',json.dumps({k:v for k,v in out.items() if k != 'ssl_context'},indent=4))
            print ('---')

    # == test defaulting to 'native'

    def test_01(self):
        self.tst0 ( ssl_opt = True , auth_opt = '' , env_opt = False )
    def test_02(self):
        self.tst0 ( ssl_opt = False, auth_opt = '' , env_opt = False )
    def test_03(self):
        self.tst0 ( ssl_opt = True , auth_opt = '' , env_opt = True )
    def test_04(self):
        self.tst0 ( ssl_opt = False, auth_opt = '' , env_opt = True  )

    # == test explicit scheme 'native'

    def test_1(self):
        self.tst0 ( ssl_opt = True , auth_opt = 'native' , env_opt = False )

    def test_2(self):
        self.tst0 ( ssl_opt = False, auth_opt = 'native' , env_opt = False )

    def test_3(self):
        self.tst0 ( ssl_opt = True , auth_opt = 'native' , env_opt = True )

    def test_4(self):
        self.tst0 ( ssl_opt = False, auth_opt = 'native' , env_opt = True  )

    # == test explicit scheme 'pam'

    def test_5(self):
        self.tst0 ( ssl_opt = True,  auth_opt = 'pam'    , env_opt = False )

    def test_6(self):
        try:
            self.tst0 ( ssl_opt = False, auth_opt = 'pam'    , env_opt = False )
        except PlainTextPAMPasswordError:
            pass
        else:
            # -- no exception raised
            self.fail("PlainTextPAMPasswordError should have been raised")

    def test_7(self):
        self.tst0 ( ssl_opt = True , auth_opt = 'pam'    , env_opt = True  )

    def test_8(self):
        self.tst0 ( ssl_opt = False, auth_opt = 'pam'    , env_opt = True  )

class TestAnonymousUser(unittest.TestCase):

    def setUp(self):
        admin = self.admin = helpers.make_session()

        user = self.user = admin.users.create('anonymous', 'rodsuser', admin.zone)
        self.home = '/{admin.zone}/home/{user.name}'.format(**locals())

        admin.collections.create(self.home)
        acl = iRODSAccess('own', self.home, user.name)
        admin.permissions.set(acl)

        self.env_file = os.path.expanduser('~/.irods.anon/irods_environment.json')
        self.env_dir = ( os.path.dirname(self.env_file))
        self.auth_file = os.path.expanduser('~/.irods.anon/.irodsA')
        os.mkdir( os.path.dirname(self.env_file))
        json.dump( { "irods_host": admin.host,
                     "irods_port": admin.port,
                     "irods_user_name": user.name,
                     "irods_zone_name": admin.zone }, open(self.env_file,'w'), indent=4 )

    def tearDown(self):
        self.admin.collections.remove(self.home, recurse = True, force = True)
        self.admin.users.remove(self.user.name)
        shutil.rmtree (self.env_dir, ignore_errors = True)

    def test_login_from_environment(self):
        orig_env = os.environ.copy()
        try:
            os.environ["IRODS_ENVIRONMENT_FILE"] = self.env_file
            os.environ["IRODS_AUTHENTICATION_FILE"] = self.auth_file
            ses = helpers.make_session()
            ses.collections.get(self.home)
        finally:
            os.environ.clear()
            os.environ.update( orig_env )

class TestWithSSL(unittest.TestCase):
    '''
    The tests within this class should be run by an account other than the
    service account.  Otherwise there is risk of corrupting the server setup.
    '''

    def setUp(self):
        if os.path.expanduser('~') == '/var/lib/irods':
            self.skipTest('TestWithSSL may not be run by user irods')
        if not os.path.exists('/etc/irods/ssl'):
            self.skipTest('Running setupssl.py as irods user is prerequisite for this test.')
        with helpers.make_session() as session:
            if not session.host in ('localhost', socket.gethostname()):
                self.skipTest('Test must be run co-resident with server')


    def test_ssl_with_server_verify_set_to_none_281(self):
        env_file = os.path.expanduser('~/.irods/irods_environment.json')
        with helpers.file_backed_up(env_file):
            with open(env_file) as env_file_handle:
                env = json.load( env_file_handle )
            env.update({ "irods_client_server_negotiation": "request_server_negotiation",
                         "irods_client_server_policy": "CS_NEG_REQUIRE",
                         "irods_ssl_ca_certificate_file": "/path/to/some/file.crt",  # does not need to exist
                         "irods_ssl_verify_server": "none",
                         "irods_encryption_key_size": 32,
                         "irods_encryption_salt_size": 8,
                         "irods_encryption_num_hash_rounds": 16,
                         "irods_encryption_algorithm": "AES-256-CBC" })
            with open(env_file,'w') as f:
                json.dump(env,f)
            with helpers.make_session() as session:
                session.collections.get('/{session.zone}/home/{session.username}'.format(**locals()))


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



from __future__ import absolute_import
import os
import json
import errno
import logging
from irods.query import Query
from irods.pool import Pool
from irods.account import iRODSAccount
from irods.manager.collection_manager import CollectionManager
from irods.manager.data_object_manager import DataObjectManager
from irods.manager.metadata_manager import MetadataManager
from irods.manager.access_manager import AccessManager
from irods.manager.user_manager import UserManager, UserGroupManager
from irods.manager.resource_manager import ResourceManager
from irods.manager.zone_manager import ZoneManager
from irods.exception import NetworkException
from irods.password_obfuscation import decode
from irods import NATIVE_AUTH_SCHEME, PAM_AUTH_SCHEME

logger = logging.getLogger(__name__)

class iRODSSession(object):

    def __init__(self, configure=True, **kwargs):
        self.pool = None
        self.numThreads = 0

        self.do_configure = (kwargs if configure else {})
        self.__configured = None
        if configure:
            self.__configured = self.configure(**kwargs)

        self.collections = CollectionManager(self)
        self.data_objects = DataObjectManager(self)
        self.metadata = MetadataManager(self)
        self.permissions = AccessManager(self)
        self.users = UserManager(self)
        self.user_groups = UserGroupManager(self)
        self.resources = ResourceManager(self)
        self.zones = ZoneManager(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cleanup()

    def __del__(self):
        self.do_configure = {}
        self.cleanup()

    def cleanup(self):
        for conn in self.pool.active | self.pool.idle:
            try:
                conn.disconnect()
            except NetworkException:
                pass
            conn.release(True)
        if self.do_configure: 
            self.__configured = self.configure(**self.do_configure)

    def _configure_account(self, **kwargs):

        try:
            env_file = kwargs['irods_env_file']

        except KeyError:
            # For backward compatibility
            for key in ['host', 'port', 'authentication_scheme']:
                if key in kwargs:
                    kwargs['irods_{}'.format(key)] = kwargs.pop(key)

            for key in ['user', 'zone']:
                if key in kwargs:
                    kwargs['irods_{}_name'.format(key)] = kwargs.pop(key)

            return iRODSAccount(**kwargs)

        # Get credentials from irods environment file
        creds = self.get_irods_env(env_file)

        # Update with new keywords arguments only
        creds.update((key, value) for key, value in kwargs.items() if key not in creds)

        # Get auth scheme
        try:
            auth_scheme = creds['irods_authentication_scheme']
        except KeyError:
            # default
            auth_scheme = 'native'

        if auth_scheme.lower() == PAM_AUTH_SCHEME:
            if 'password' in creds:
                return iRODSAccount(**creds)
            else:
                # password will be from irodsA file therefore use native login
                creds['irods_authentication_scheme'] = NATIVE_AUTH_SCHEME
        elif auth_scheme != 'native':
            return iRODSAccount(**creds)

        # Native auth, try to unscramble password
        try:
            creds['irods_authentication_uid'] = kwargs['irods_authentication_uid']
        except KeyError:
            pass

        creds['password'] = self.get_irods_password(**creds)

        return iRODSAccount(**creds)

    def configure(self, **kwargs):
        account = self.__configured
        if not account:
            account = self._configure_account(**kwargs)
        connection_refresh_time = self.get_connection_refresh_time(**kwargs)
        logger.debug("In iRODSSession's configure(). connection_refresh_time set to {}".format(connection_refresh_time))
        self.pool = Pool(account, application_name=kwargs.pop('application_name',''), connection_refresh_time=connection_refresh_time)
        return account

    def query(self, *args):
        return Query(self, *args)

    @property
    def username(self):
        return self.pool.account.client_user

    @property
    def zone(self):
        return self.pool.account.client_zone

    @property
    def host(self):
        return self.pool.account.host

    @property
    def port(self):
        return self.pool.account.port

    @property
    def server_version(self):
        try:
            conn = next(iter(self.pool.active))
            return conn.server_version
        except StopIteration:
            conn = self.pool.get_connection()
            version = conn.server_version
            conn.release()
            return version

    @property
    def pam_pw_negotiated(self):
            self.pool.account.store_pw = []
            conn = self.pool.get_connection()
            pw = getattr(self.pool.account,'store_pw',[])
            delattr( self.pool.account, 'store_pw')
            conn.release()
            return pw

    @property
    def default_resource(self):
        return self.pool.account.default_resource

    @default_resource.setter
    def default_resource(self, name):
        self.pool.account.default_resource = name

    @property
    def connection_timeout(self):
        return self.pool.connection_timeout

    @connection_timeout.setter
    def connection_timeout(self, seconds):
        self.pool.connection_timeout = seconds

    @staticmethod
    def get_irods_password_file():
        try:
            return os.environ['IRODS_AUTHENTICATION_FILE']
        except KeyError:
            return os.path.expanduser('~/.irods/.irodsA')

    @staticmethod
    def get_irods_env(env_file):
        try:
            with open(env_file, 'rt') as f:
                return json.load(f)
        except IOError:
            logger.debug("Could not open file {}".format(env_file))
            return {}

    @staticmethod
    def get_irods_password(**kwargs):
        try:
            irods_auth_file = kwargs['irods_authentication_file']
        except KeyError:
            irods_auth_file = iRODSSession.get_irods_password_file()

        try:
            uid = kwargs['irods_authentication_uid']
        except KeyError:
            uid = None

        try:
            with open(irods_auth_file, 'r') as f:
                return decode(f.read().rstrip('\n'), uid)
        except IOError as exc:
            if exc.errno != errno.ENOENT: raise # Auth file exists but can't be read
            return ''                           # No auth file (as with anonymous user)

    def get_connection_refresh_time(self, **kwargs):
        connection_refresh_time = -1
        
        connection_refresh_time = int(kwargs.get('refresh_time', -1))
        if connection_refresh_time != -1:
            return connection_refresh_time

        try:
            env_file = kwargs['irods_env_file']
        except KeyError:
            return connection_refresh_time

        if env_file is not None:
            env_file_map = self.get_irods_env(env_file)
            connection_refresh_time = int(env_file_map.get('irods_connection_refresh_time', -1))
            if connection_refresh_time < 1:
                # Negative values are not allowed.
                logger.debug('connection_refresh_time in {} file has value of {}. Only values greater than 1 are allowed.'.format(env_file, connection_refresh_time))
                connection_refresh_time = -1

        return connection_refresh_time



#! /usr/bin/env python
from __future__ import absolute_import
import os
import sys
import unittest
from irods.exception import UserGroupDoesNotExist
from irods.meta import iRODSMetaCollection, iRODSMeta
from irods.models import User, UserGroup, UserMeta
import irods.test.helpers as helpers
from six.moves import range


class TestUserGroup(unittest.TestCase):

    def setUp(self):
        self.sess = helpers.make_session()

    def tearDown(self):
        '''Close connections
        '''
        self.sess.cleanup()

    def test_create_group(self):
        group_name = "test_group"

        # group should not be already present
        with self.assertRaises(UserGroupDoesNotExist):
            self.sess.user_groups.get(group_name)

        # create group
        group = self.sess.user_groups.create(group_name)

        # assertions
        self.assertEqual(group.name, group_name)
        self.assertEqual(
            repr(group), "<iRODSUserGroup {0} {1}>".format(group.id, group_name))

        # delete group
        group.remove()

        # group should be gone
        with self.assertRaises(UserGroupDoesNotExist):
            self.sess.user_groups.get(group_name)

    def test_add_users_to_group(self):
        group_name = "test_group"
        group_size = 15
        base_user_name = "test_user"

        # group should not be already present
        with self.assertRaises(UserGroupDoesNotExist):
            self.sess.user_groups.get(group_name)

        # create test group
        group = self.sess.user_groups.create(group_name)

        # create test users names
        test_user_names = []
        for i in range(group_size):
            test_user_names.append(base_user_name + str(i))

        # make test users and add them to test group
        for test_user_name in test_user_names:
            user = self.sess.users.create(test_user_name, 'rodsuser')
            group.addmember(user.name)

        # list group members
        member_names = [user.name for user in group.members]

        # compare lists
        self.assertSetEqual(set(member_names), set(test_user_names))

        # exercise iRODSUserGroup.hasmember()
        for test_user_name in test_user_names:
            self.assertTrue(group.hasmember(test_user_name))

        # remove test users from group and delete them
        for test_user_name in test_user_names:
            user = self.sess.users.get(test_user_name)
            group.removemember(user.name)
            user.remove()

        # delete group
        group.remove()

        # group should be gone
        with self.assertRaises(UserGroupDoesNotExist):
            self.sess.user_groups.get(group_name)

    def test_user_dn(self):
        # https://github.com/irods/irods/issues/3620
        if self.sess.server_version == (4, 2, 1):
            self.skipTest('Broken in 4.2.1')

        user_name = "testuser"
        user_DNs = ['foo', 'bar']

        # create user
        user = self.sess.users.create(user_name, 'rodsuser')

        # expect no dn
        self.assertEqual(user.dn, [])

        # add dn
        user.modify('addAuth', user_DNs[0])
        self.assertEqual(user.dn[0], user_DNs[0])

        # add other dn
        user.modify('addAuth', user_DNs[1])
        self.assertEqual( sorted(user.dn), sorted(user_DNs) )

        # remove first dn
        user.modify('rmAuth', user_DNs[0])

        # confirm removal
        self.assertEqual(sorted(user.dn), sorted(user_DNs[1:]))

        # delete user
        user.remove()

    def test_group_metadata(self):
        group_name = "test_group"

        # group should not be already present
        with self.assertRaises(UserGroupDoesNotExist):
            self.sess.user_groups.get(group_name)

        group = None

        try:
            # create group
            group = self.sess.user_groups.create(group_name)

            # add metadata to group
            triple = ['key', 'value', 'unit']
            group.metadata[triple[0]] = iRODSMeta(*triple)

            result =  self.sess.query(UserMeta, UserGroup).filter(UserGroup.name == group_name,
                                                                  UserMeta.name == 'key').one()

            self.assertTrue([result[k] for k in (UserMeta.name, UserMeta.value, UserMeta.units)] == triple)

        finally:
            if group:
                group.remove()
                helpers.remove_unused_metadata(self.sess)

    def test_user_metadata(self):
        user_name = "testuser"
        user = None

        try:
            user = self.sess.users.create(user_name, 'rodsuser')

            # metadata collection is the right type?
            self.assertIsInstance(user.metadata, iRODSMetaCollection)

            # add three AVUs, two having the same key
            user.metadata['key0'] = iRODSMeta('key0', 'value', 'units')
            sorted_triples = sorted( [ ['key1', 'value0', 'units0'],
                                       ['key1', 'value1', 'units1']  ] )
            for m in sorted_triples:
                user.metadata.add(iRODSMeta(*m))

            # general query gives the right results?
            result_0 =  self.sess.query(UserMeta, User)\
                         .filter( User.name == user_name, UserMeta.name == 'key0').one()

            self.assertTrue( [result_0[k] for k in (UserMeta.name, UserMeta.value, UserMeta.units)]
                              == ['key0', 'value', 'units'] )

            results_1 =  self.sess.query(UserMeta, User)\
                         .filter(User.name == user_name, UserMeta.name == 'key1')

            retrieved_triples = [ [ res[k] for k in (UserMeta.name, UserMeta.value, UserMeta.units) ]
                                  for res in results_1
                                ]

            self.assertTrue( sorted_triples == sorted(retrieved_triples))

        finally:
            if user:
                user.remove()
                helpers.remove_unused_metadata(self.sess)

    def test_get_user_metadata(self):
        user_name = "testuser"
        user = None

        try:
            # create user
            user = self.sess.users.create(user_name, 'rodsuser')
            meta = user.metadata.get_all('key')

            # There should be no metadata
            self.assertEqual(len(meta), 0)
        finally:
            if user: user.remove()

    def test_add_user_metadata(self):
        user_name = "testuser"
        user = None

        try:
            # create user
            user = self.sess.users.create(user_name, 'rodsuser')

            user.metadata.add('key0', 'value0')
            user.metadata.add('key1', 'value1', 'unit1')
            user.metadata.add('key2', 'value2a', 'unit2')
            user.metadata.add('key2', 'value2b', 'unit2')

            meta0 = user.metadata.get_all('key0')
            self.assertEqual(len(meta0),1)
            self.assertEqual(meta0[0].name, 'key0')
            self.assertEqual(meta0[0].value, 'value0')

            meta1 = user.metadata.get_all('key1')
            self.assertEqual(len(meta1),1)
            self.assertEqual(meta1[0].name, 'key1')
            self.assertEqual(meta1[0].value, 'value1')
            self.assertEqual(meta1[0].units, 'unit1')

            meta2 = sorted(user.metadata.get_all('key2'), key = lambda AVU : AVU.value)
            self.assertEqual(len(meta2),2)
            self.assertEqual(meta2[0].name, 'key2')
            self.assertEqual(meta2[0].value, 'value2a')
            self.assertEqual(meta2[0].units, 'unit2')
            self.assertEqual(meta2[1].name, 'key2')
            self.assertEqual(meta2[1].value, 'value2b')
            self.assertEqual(meta2[1].units, 'unit2')

            user.metadata.remove('key1', 'value1', 'unit1')
            metadata = user.metadata.items()
            self.assertEqual(len(metadata), 3)

            user.metadata.remove('key2', 'value2a', 'unit2')
            metadata = user.metadata.items()
            self.assertEqual(len(metadata), 2)

        finally:
            if user:
                user.remove()
                helpers.remove_unused_metadata(self.sess)

if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import sys
import time
import textwrap
import unittest
from irods.models import DataObject
from irods.exception import (FAIL_ACTION_ENCOUNTERED_ERR, RULE_ENGINE_ERROR, UnknowniRODSError)
import irods.test.helpers as helpers
from irods.rule import Rule
import six
from io import open as io_open
import io


RE_Plugins_installed_run_condition_args = ( os.environ.get('PYTHON_RULE_ENGINE_INSTALLED','*').lower()[:1]=='y',
                                           'Test depends on server having Python-REP installed beyond the default options' )


class TestRule(unittest.TestCase):

    '''Suite of tests on rule operations
    '''

    def setUp(self):
        self.sess = helpers.make_session()

    def tearDown(self):
        # close connections
        self.sess.cleanup()

    def test_add_metadata_from_rule_file(self):
        '''
        Tests running a rule from a client-side .r file.
        The rule adds metadata attributes to an iRODS object
        and we check for the presence of said attributes.
        '''
        session = self.sess

        # test metadata
        attr_name = "test_attr"
        attr_value = "test_value"

        # make test object
        ts = time.time()
        zone = session.zone
        username = session.username
        object_name = 'foo_{ts}.txt'.format(**locals())

        object_path = "/{zone}/home/{username}/{object_name}".format(
            **locals())
        obj = helpers.make_object(session, object_path)

        # make rule file
        rule_file_path = "/tmp/test_{ts}.r".format(**locals())
        rule = textwrap.dedent('''\
                                test {{
                                    # add metadata
                                    *attribute.*name = *value;
                                    msiAssociateKeyValuePairsToObj(*attribute, *object, "-d")
                                }}
                                INPUT *object="{object_path}",*name="{attr_name}",*value="{attr_value}"
                                OUTPUT ruleExecOut'''.format(**locals()))

        with open(rule_file_path, "w") as rule_file:
            rule_file.write(rule)

        # run test rule
        myrule = Rule(session, rule_file_path)
        myrule.execute()

        # check that metadata is there
        meta = session.metadata.get(DataObject, object_path)
        assert meta[0].name == attr_name
        assert meta[0].value == attr_value

        # remove test object
        obj.unlink(force=True)

        # remove rule file
        os.remove(rule_file_path)

    def test_set_metadata_288(self):

        session = self.sess

        # rule body
        rule_body = textwrap.dedent('''\
                                    *attribute.*attr_name = "*attr_value"
                                    msiAssociateKeyValuePairsToObj(*attribute, *object, "-d")
                                    # (: -- just a comment -- :)  writeLine("serverLog","*value")
                                    ''')

        input_params = { '*value': "3334" , "*object": '/tempZone/home/rods/runner.py' ,
                                          '*attr_name':'XX',
                                          '*attr_value':'YY'
        }

        output = 'ruleExecOut'

        myrule = Rule(session, body=rule_body, params=input_params, output=output)
        myrule.execute()


    # test catching fail-type actions initiated directly in the instance being called.
    #
    @unittest.skipUnless (*RE_Plugins_installed_run_condition_args)
    def test_with_fail_in_targeted_rule_engines(self):
        self._failing_in_targeted_rule_engines(rule_to_call = "generic_failing_rule")


    # test catching rule fail actions initiated using the native 'failmsg' microservice.
    #
    @unittest.skipUnless (*RE_Plugins_installed_run_condition_args)
    def test_with_using_native_fail_msvc(self):
        error_dict = \
        self._failing_in_targeted_rule_engines(rule_to_call = [('irods_rule_engine_plugin-python-instance','failing_with_message_py'),
                                                               ('irods_rule_engine_plugin-irods_rule_language-instance','failing_with_message')])
        for v in error_dict.values():
            self.assertIn('code of minus 2', v[1].args[0])

    # helper for the previous two tests.
    #
    def _failing_in_targeted_rule_engines(self, rule_to_call = None):
        session = self.sess
        if isinstance(rule_to_call,(list,tuple)):
            rule_dict = dict(rule_to_call)
        else:
            rule_dict = {}

        rule_instances_list = ( 'irods_rule_engine_plugin-irods_rule_language-instance',
                                'irods_rule_engine_plugin-python-instance' )
        err_hash = {}

        for i in rule_instances_list:

            if rule_dict:
                rule_to_call = rule_dict[i]

            rule = Rule( session, body = rule_to_call,
                         instance_name = i )
            try:
                rule.execute( acceptable_errors = (-1,) )
            except UnknowniRODSError as e:
                err_hash[i] = ('rule exec failed! - misc - ',(e)) # 2-tuple = failure
            except RULE_ENGINE_ERROR as e:
                err_hash[i] = ('rule exec failed! - python - ',(e)) # 2-tuple = failure
            except FAIL_ACTION_ENCOUNTERED_ERR as e:
                err_hash[i] = ('rule exec failed! - native - ',(e))
            else:
                err_hash[i] = ('rule exec succeeded!',) # 1-tuple = success

        self.assertEqual( len(err_hash), len(rule_instances_list) )
        self.assertEqual( len(err_hash), len([val for val in err_hash.values() if val[0].startswith('rule exec failed')]) )
        return err_hash


    @unittest.skipUnless (*RE_Plugins_installed_run_condition_args)
    def test_targeting_Python_instance_when_rule_multiply_defined(self):
        self._with_X_instance_when_rule_multiply_defined(
            instance_name  = 'irods_rule_engine_plugin-python-instance',
            test_condition = lambda bstring: b'python' in bstring
            )

    @unittest.skipUnless (*RE_Plugins_installed_run_condition_args)
    def test_targeting_Native_instance_when_rule_multiply_defined(self):
        self._with_X_instance_when_rule_multiply_defined(
            instance_name  = 'irods_rule_engine_plugin-irods_rule_language-instance',
            test_condition = lambda bstring: b'native' in bstring
            )

    @unittest.skipUnless (*RE_Plugins_installed_run_condition_args)
    def test_targeting_Unspecified_instance_when_rule_multiply_defined(self):
        self._with_X_instance_when_rule_multiply_defined(
            test_condition = lambda bstring: b'native' in bstring and b'python' in bstring
            )

    def _with_X_instance_when_rule_multiply_defined(self,**kw):
        session = self.sess
        rule = Rule( session, body = 'defined_in_both',
                     output = 'ruleExecOut',
                     **{key:val for key,val in kw.items() if key == 'instance_name'}
                   )
        output  = rule.execute()
        buf = output.MsParam_PI[0].inOutStruct.stdoutBuf.buf
        self.assertTrue(kw['test_condition'](buf.rstrip(b'\0').rstrip()))


    def test_specifying_rule_instance(self):

        self._with_writeline_to_stream(
                stream_name = 'stdout',
                rule_engine_instance = "irods_rule_engine_plugin-irods_rule_language-instance" )


    def _with_writeline_to_stream(self, stream_name = "serverLog",
                                          output_string = 'test-writeline-to-stream',
                                          alternate_input_params = (),
                                          rule_engine_instance = ""):

        session = self.sess

        # rule body
        rule_body = textwrap.dedent('''\
                                    writeLine("{stream_name}","*value")
                                    '''.format(**locals()))

        input_params = { '*value': output_string }
        input_params.update( alternate_input_params )

        output_param = 'ruleExecOut'

        extra_options = {}

        if rule_engine_instance:
            extra_options [ 'instance_name' ] = rule_engine_instance

        myrule = Rule(session, body=rule_body, params=input_params, output=output_param, **extra_options)
        output = myrule.execute()

        buf = None
        if stream_name == 'stdout':
            buf = output.MsParam_PI[0].inOutStruct.stdoutBuf.buf
        elif stream_name == 'stderr':
            buf = output.MsParam_PI[0].inOutStruct.stderrBuf.buf

        if buf is not None:
            buf = buf.decode('utf-8')
            self.assertEqual (output_string, buf.rstrip('\0').rstrip())


    def test_add_metadata_from_rule(self):
        '''
        Runs a rule whose body and input parameters are created in our script.
        The rule adds metadata attributes to an iRODS object
        and we check for the presence of said attributes.
        '''
        session = self.sess

        # test metadata
        attr_name = "test_attr"
        attr_value = "test_value"

        # make test object
        ts = time.time()
        zone = session.zone
        username = session.username
        object_name = 'foo_{ts}.txt'.format(**locals())

        object_path = "/{zone}/home/{username}/{object_name}".format(
            **locals())
        obj = helpers.make_object(session, object_path)

        # rule body
        rule_body = textwrap.dedent('''\
                                test {{
                                    # add metadata
                                    *attribute.*name = *value;
                                    msiAssociateKeyValuePairsToObj(*attribute, *object, "-d")
                                }}''')

        # rule parameters
        input_params = {  # extra quotes for string literals
            '*object': '"{object_path}"'.format(**locals()),
            '*name': '"{attr_name}"'.format(**locals()),
            '*value': '"{attr_value}"'.format(**locals())
        }
        output = 'ruleExecOut'

        # run test rule
        myrule = Rule(session, body=rule_body, irods_3_literal_style = True,
                      params=input_params, output=output)
        myrule.execute()

        # check that metadata is there
        meta = session.metadata.get(DataObject, object_path)
        assert meta[0].name == attr_name
        assert meta[0].value == attr_value

        # remove test object
        obj.unlink(force=True)


    def test_retrieve_std_streams_from_rule(self):
        '''
        Tests running a rule from a client-side .r file.
        The rule writes things to its stdout that we
        get back on the client side
        '''

        # Wrong buffer length on older versions
        if self.sess.server_version < (4, 1, 7):
            self.skipTest('For iRODS 4.1.7 and newer')

        session = self.sess

        # test metadata
        some_string = u'foo'
        some_other_string = u'我喜欢麦当劳'
        err_string = u'⛔'

        # make rule file
        ts = time.time()
        rule_file_path = "/tmp/test_{ts}.r".format(**locals())
        rule = textwrap.dedent(u'''\
                                test {{
                                    # write stuff
                                    writeLine("stdout", *some_string);
                                    writeLine("stdout", *some_other_string);
                                    writeLine("stderr", *err_string);
                                }}
                                INPUT *some_string="{some_string}",*some_other_string="{some_other_string}",*err_string="{err_string}"
                                OUTPUT ruleExecOut'''.format(**locals()))

        with io_open(rule_file_path, "w", encoding='utf-8') as rule_file:
            rule_file.write(rule)

        # run test rule
        myrule = Rule(session, rule_file_path)
        out_array = myrule.execute()

        # retrieve out buffer
        buf = out_array.MsParam_PI[0].inOutStruct.stdoutBuf.buf

        # it's binary data (BinBytesBuf) so must be decoded
        buf = buf.decode('utf-8')

        # check that we get our strings back
        self.assertIn(some_string, buf)
        self.assertIn(some_other_string, buf)

        # same thing stderr buffer
        buf = out_array.MsParam_PI[0].inOutStruct.stderrBuf.buf

        # decode and check
        buf = buf.decode('utf-8')
        self.assertIn(err_string, buf)

        # remove rule file
        os.remove(rule_file_path)


    @staticmethod
    def lines_from_stdout_buf(output):
        buf = ""
        if output and len(output.MsParam_PI):
            buf = output.MsParam_PI[0].inOutStruct.stdoutBuf.buf
            if buf:
                buf = buf.rstrip(b'\0').decode('utf8')
        return buf.splitlines()


    def test_rulefile_in_file_like_object_1__336(self):

        rule_file_contents = textwrap.dedent(u"""\
        hw {
                helloWorld(*message);
                writeLine("stdout", "Message is: [*message] ...");
        }
        helloWorld(*OUT)
        {
          *OUT = "Hello world!"
        }
        """)
        r = Rule(self.sess, rule_file = io.StringIO( rule_file_contents ),
                            output = 'ruleExecOut', instance_name='irods_rule_engine_plugin-irods_rule_language-instance')
        output = r.execute()
        lines = self.lines_from_stdout_buf(output)
        self.assertRegexpMatches (lines[0], '.*\[Hello world!\]')


    def test_rulefile_in_file_like_object_2__336(self):

        rule_file_contents = textwrap.dedent("""\
        main {
          other_rule()
          writeLine("stdout","["++type(*msg2)++"][*msg2]");
        }
        other_rule {
          writeLine("stdout","["++type(*msg1)++"][*msg1]");
        }

        INPUT *msg1="",*msg2=""
        OUTPUT ruleExecOut
        """)

        r = Rule(self.sess, rule_file = io.BytesIO( rule_file_contents.encode('utf-8') ))
        output = r.execute()
        lines = self.lines_from_stdout_buf(output)
        self.assertRegexpMatches (lines[0], '\[STRING\]\[\]')
        self.assertRegexpMatches (lines[1], '\[STRING\]\[\]')

        r = Rule(self.sess, rule_file = io.BytesIO( rule_file_contents.encode('utf-8') )
                          , params = {'*msg1':5, '*msg2':'"A String"'})
        output = r.execute()
        lines = self.lines_from_stdout_buf(output)
        self.assertRegexpMatches (lines[0], '\[INTEGER\]\[5\]')
        self.assertRegexpMatches (lines[1], '\[STRING\]\[A String\]')


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



from __future__ import absolute_import
import logging

from irods.models import User, UserGroup
from irods.manager import Manager
from irods.message import GeneralAdminRequest, iRODSMessage, GetTempPasswordForOtherRequest, GetTempPasswordForOtherOut
from irods.exception import UserDoesNotExist, UserGroupDoesNotExist, NoResultFound, CAT_SQL_ERR
from irods.api_number import api_number
from irods.user import iRODSUser, iRODSUserGroup
import irods.password_obfuscation as obf

logger = logging.getLogger(__name__)


class UserManager(Manager):

    def get(self, user_name, user_zone=""):
        query = self.sess.query(User).filter(User.name == user_name)

        if len(user_zone) > 0:
            query = query.filter(User.zone == user_zone)

        try:
            result = query.one()
        except NoResultFound:
            raise UserDoesNotExist()
        return iRODSUser(self, result)

    def create(self, user_name, user_type, user_zone="", auth_str=""):
        message_body = GeneralAdminRequest(
            "add",
            "user",
            user_name if not user_zone or user_zone == self.sess.zone \
                      else "{}#{}".format(user_name,user_zone),
            user_type,
            user_zone,
            auth_str
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)
        return self.get(user_name, user_zone)

    def remove(self, user_name, user_zone=""):
        message_body = GeneralAdminRequest(
            "rm",
            "user",
            user_name,
            user_zone
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    def temp_password_for_user(self, user_name):
        with self.sess.pool.get_connection() as conn:
            message_body = GetTempPasswordForOtherRequest(
                targetUser=user_name,
                unused=None
            )
            request = iRODSMessage("RODS_API_REQ", msg=message_body,
                                   int_info=api_number['GET_TEMP_PASSWORD_FOR_OTHER_AN'])

            # Send request
            conn.send(request)

            # Receive answer
            try:
                response = conn.recv()
                logger.debug(response.int_info)
            except CAT_SQL_ERR:
                raise UserDoesNotExist()

            # Convert and return answer
            msg = response.get_main_message(GetTempPasswordForOtherOut)
            return obf.create_temp_password(msg.stringToHashWith, conn.account.password)


    def modify(self, user_name, option, new_value, user_zone=""):

        # must append zone to username for this API call
        if len(user_zone) > 0:
            user_name += "#" + user_zone

        with self.sess.pool.get_connection() as conn:

            # if modifying password, new value needs obfuscating
            if option == 'password':
                current_password = self.sess.pool.account.password
                new_value = obf.obfuscate_new_password(new_value, current_password, conn.client_signature)

            message_body = GeneralAdminRequest(
                "modify",
                "user",
                user_name,
                option,
                new_value,
                user_zone,
            )
            request = iRODSMessage("RODS_API_REQ", msg=message_body,
                                   int_info=api_number['GENERAL_ADMIN_AN'])

            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)


class UserGroupManager(UserManager):

    def get(self, name, user_zone=""):
        query = self.sess.query(UserGroup).filter(UserGroup.name == name)

        try:
            result = query.one()
        except NoResultFound:
            raise UserGroupDoesNotExist()
        return iRODSUserGroup(self, result)

    def create(self, name, user_type='rodsgroup', user_zone="", auth_str=""):
        message_body = GeneralAdminRequest(
            "add",
            "user",
            name,
            user_type,
            "",
            ""
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)
        return self.get(name)

    def getmembers(self, name):
        results = self.sess.query(User).filter(
            User.type != 'rodsgroup', UserGroup.name == name)
        return [iRODSUser(self, row) for row in results]

    def addmember(self, group_name, user_name, user_zone=""):
        message_body = GeneralAdminRequest(
            "modify",
            "group",
            group_name,
            "add",
            user_name,
            user_zone
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)

    def removemember(self, group_name, user_name, user_zone=""):
        message_body = GeneralAdminRequest(
            "modify",
            "group",
            group_name,
            "remove",
            user_name,
            user_zone
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])
        with self.sess.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        logger.debug(response.int_info)



from __future__ import absolute_import
import os
import io
import tempfile
import contextlib
import shutil
import hashlib
import base64
import math
import socket
import inspect
import threading
import random
import datetime
from pwd import getpwnam
from irods.session import iRODSSession
from irods.message import iRODSMessage
from six.moves import range


def my_function_name():
    """Returns the name of the calling function or method"""
    return inspect.getframeinfo(inspect.currentframe().f_back).function


_thrlocal = threading.local()

def unique_name(*seed_tuple):
    '''For deterministic pseudo-random identifiers based on function/method name
       to prevent e.g.  ICAT collisions within and between tests.  Example use:

           def f(session):
             seq_num = 1
             a_name = unique_name( my_function_name(), seq_num # [, *optional_further_args]
                                  )
             seq_num += 1
             session.resources.create( a_name, 'unixfilesystem', session.host, '/tmp/' + a_name )
    '''
    if not getattr(_thrlocal,"rand_gen",None) : _thrlocal.rand_gen = random.Random()
    _thrlocal.rand_gen.seed(seed_tuple)
    return '%016X' % _thrlocal.rand_gen.randint(0,(1<<64)-1)


IRODS_SHARED_DIR = os.path.join( os.path.sep, 'irods_shared' )
IRODS_SHARED_TMP_DIR = os.path.join(IRODS_SHARED_DIR,'tmp')
IRODS_SHARED_REG_RESC_VAULT = os.path.join(IRODS_SHARED_DIR,'reg_resc')

IRODS_REG_RESC = 'MyRegResc'

def irods_shared_tmp_dir():
    pth = IRODS_SHARED_TMP_DIR
    can_write = False
    if os.path.exists(pth):
        try:     tempfile.NamedTemporaryFile(dir = pth)
        except:  pass
        else:    can_write = True 
    return pth if can_write else ''

def irods_shared_reg_resc_vault() :
    vault = IRODS_SHARED_REG_RESC_VAULT
    if os.path.exists(vault):
        return vault
    else:
        return None

def get_register_resource(session):
    vault_path = irods_shared_reg_resc_vault()
    Reg_Resc_Name = ''
    if vault_path:
        session.resources.create(IRODS_REG_RESC, 'unixfilesystem', session.host, vault_path)
        Reg_Resc_Name = IRODS_REG_RESC
    return Reg_Resc_Name


def make_session(**kwargs):
    try:
        env_file = kwargs.pop('irods_env_file')
    except KeyError:
        try:
            env_file = os.environ['IRODS_ENVIRONMENT_FILE']
        except KeyError:
            env_file = os.path.expanduser('~/.irods/irods_environment.json')

    try:
        os.environ['IRODS_CI_TEST_RUN']
        uid = getpwnam('irods').pw_uid
    except KeyError:
        uid = None

    return iRODSSession( irods_authentication_uid = uid, irods_env_file = env_file, **kwargs )


def home_collection(session):
    return "/{0.zone}/home/{0.username}".format(session)


def make_object(session, path, content=None, **options):
    if content is None:
        content = u'blah'

    content = iRODSMessage.encode_unicode(content)

    if session.server_version <= (4,1,4):
        # 2 step open-create necessary for iRODS 4.1.4 or older
        obj = session.data_objects.create(path)
        with obj.open('w', **options) as obj_desc:
            obj_desc.write(content)
    else:
        with session.data_objects.open(path, 'w', **options) as obj_desc:
            obj_desc.write(content)

    # refresh object after write
    return session.data_objects.get(path)


def make_collection(session, path, object_names=None, object_content=None):
    # create collection
    coll = session.collections.create(path)

    # create objects
    if object_names:
        for name in object_names:
            obj_path = os.path.join(path, name)
            make_object(session, obj_path, content=object_content)

    return coll


def make_test_collection(session, path, obj_count):
    coll = session.collections.create(path)

    for n in range(obj_count):
        obj_path = path + "/test" + str(n).zfill(6) + ".txt"
        make_object(session, obj_path)

    return coll


def make_deep_collection(session, root_path, depth=10, objects_per_level=50, object_content=None):
    # start at root path
    current_coll_path = root_path

    # make collections recursively
    for d in range(depth):
        # make list of object names
        obj_names = ['obj' + str(i).zfill(len(str(objects_per_level)))
                     for i in range(objects_per_level)]

        # make subcollection and objects
        if d == 0:
            root_coll = make_collection(
                session, current_coll_path, obj_names, object_content)
        else:
            make_collection(
                session, current_coll_path, obj_names, object_content)

        # next level down
        current_coll_path = os.path.join(
            current_coll_path, 'subcoll' + str(d).zfill(len(str(d))))

    return root_coll


def make_flat_test_dir(dir_path, file_count=10, file_size=1024):
    if file_count < 1:
        raise ValueError

    os.mkdir(dir_path)

    for i in range(file_count):
        # pad file name suffix with zeroes
        suffix_width = int(math.log10(file_count))+1
        file_path = '{dir_path}/test_{i:0>{suffix_width}}.txt'.format(**locals())

        # make random binary file
        with open(file_path, 'wb') as f:
            f.write(os.urandom(file_size))

@contextlib.contextmanager
def create_simple_resc (self, rescName = None):
    if not rescName: 
        rescName =  'simple_resc_' + unique_name (my_function_name() + '_simple_resc', datetime.datetime.now())
    created = False
    try:
        self.sess.resources.create(rescName,
                                   'unixfilesystem',
                                   host = self.sess.host,
                                   path = '/tmp/' + rescName)
        created = True
        yield rescName
    finally:
        if created:
            self.sess.resources.remove(rescName)

@contextlib.contextmanager
def create_simple_resc_hierarchy (self, Root, Leaf):
    d = tempfile.mkdtemp()
    self.sess.resources.create(Leaf,'unixfilesystem',
                           host = self.sess.host,
                           path=d)
    self.sess.resources.create(Root,'passthru')
    self.sess.resources.add_child(Root,Leaf)
    try:
        yield ';'.join([Root,Leaf])
    finally:
        self.sess.resources.remove_child(Root,Leaf)
        self.sess.resources.remove(Leaf)
        self.sess.resources.remove(Root)
        shutil.rmtree(d)


def chunks(f, chunksize=io.DEFAULT_BUFFER_SIZE):
    return iter(lambda: f.read(chunksize), b'')


def compute_sha256_digest(file_path):
    hasher = hashlib.sha256()

    with open(file_path, 'rb') as f:
        for chunk in chunks(f):
            hasher.update(chunk)

    return base64.b64encode(hasher.digest()).decode()


def remove_unused_metadata(session):
    from irods.message import GeneralAdminRequest
    from irods.api_number import api_number
    message_body = GeneralAdminRequest( 'rm', 'unusedAVUs', '','','','')
    req = iRODSMessage("RODS_API_REQ", msg = message_body,int_info=api_number['GENERAL_ADMIN_AN'])
    with session.pool.get_connection() as conn:
        conn.send(req)
        response=conn.recv()
        if (response.int_info != 0): raise RuntimeError("Error removing unused AVUs")


@contextlib.contextmanager
def file_backed_up(filename):
    with tempfile.NamedTemporaryFile(prefix=os.path.basename(filename)) as f:
        shutil.copyfile(filename, f.name)
        try:
            yield filename
        finally:
            shutil.copyfile(f.name, filename)


def irods_session_host_local (sess):
    return socket.gethostbyname(sess.host) == \
           socket.gethostbyname(socket.gethostname())



from __future__ import absolute_import
import socket
import logging
import struct
import hashlib
import six
import os
import ssl
import datetime
import irods.password_obfuscation as obf
from irods import MAX_NAME_LEN
from ast import literal_eval as safe_eval


from irods.message import (
    iRODSMessage, StartupPack, AuthResponse, AuthChallenge, AuthPluginOut,
    OpenedDataObjRequest, FileSeekResponse, StringStringMap, VersionResponse,
    PluginAuthMessage, ClientServerNegotiation, Error, GetTempPasswordOut)
from irods.exception import (get_exception_by_code, NetworkException, nominal_code)
from irods.message import (PamAuthRequest, PamAuthRequestOut)


ALLOW_PAM_LONG_TOKENS = True      # True to fix [#279]


from irods import (
    MAX_PASSWORD_LENGTH, RESPONSE_LEN,
    AUTH_SCHEME_KEY, AUTH_USER_KEY, AUTH_PWD_KEY, AUTH_TTL_KEY,
    NATIVE_AUTH_SCHEME,
    GSI_AUTH_PLUGIN, GSI_AUTH_SCHEME, GSI_OID,
    PAM_AUTH_SCHEME)
from irods.client_server_negotiation import (
    perform_negotiation,
    validate_policy,
    REQUEST_NEGOTIATION,
    REQUIRE_TCP,
    FAILURE,
    USE_SSL,
    CS_NEG_RESULT_KW)
from irods.api_number import api_number

logger = logging.getLogger(__name__)

class PlainTextPAMPasswordError(Exception): pass

class Connection(object):

    DISALLOWING_PAM_PLAINTEXT = True

    def __init__(self, pool, account):

        self.pool = pool
        self.socket = None
        self.account = account
        self._client_signature = None
        self._server_version = self._connect()
        self._disconnected = False

        scheme = self.account.authentication_scheme

        if scheme == NATIVE_AUTH_SCHEME:
            self._login_native()
        elif scheme == GSI_AUTH_SCHEME:
            self.client_ctx = None
            self._login_gsi()
        elif scheme == PAM_AUTH_SCHEME:
            self._login_pam()
        else:
            raise ValueError("Unknown authentication scheme %s" % scheme)
        self.create_time = datetime.datetime.now()
        self.last_used_time = self.create_time

    @property
    def server_version(self):
        detected = tuple(int(x) for x in self._server_version.relVersion.replace('rods', '').split('.'))
        return (safe_eval(os.environ.get('IRODS_SERVER_VERSION','()'))
                or detected)
    @property
    def client_signature(self):
        return self._client_signature

    def __del__(self):
        if self.socket and getattr(self,"_disconnected",False):
            self.disconnect()

    def send(self, message):
        string = message.pack()

        logger.debug(string)
        try:
            self.socket.sendall(string)
        except:
            logger.error(
                "Unable to send message. " +
                "Connection to remote host may have closed. " +
                "Releasing connection from pool."
            )
            self.release(True)
            raise NetworkException("Unable to send message")

    def recv(self, into_buffer = None
                 , return_message = ()
                 , acceptable_errors = ()):
        acceptable_codes = set(nominal_code(e) for e in acceptable_errors)
        try:
            if into_buffer is None:
                msg = iRODSMessage.recv(self.socket)
            else:
                msg = iRODSMessage.recv_into(self.socket, into_buffer)
        except socket.error:
            logger.error("Could not receive server response")
            self.release(True)
            raise NetworkException("Could not receive server response")
        if isinstance(return_message,list): return_message[:] = [msg]
        if msg.int_info < 0:
            try:
                err_msg = iRODSMessage(msg=msg.error).get_main_message(Error).RErrMsg_PI[0].msg
            except TypeError:
                err_msg = None
            if nominal_code(msg.int_info) not in acceptable_codes:
                raise get_exception_by_code(msg.int_info, err_msg)
        return msg

    def recv_into(self, buffer, **options):
        return self.recv( into_buffer = buffer, **options )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def release(self, destroy=False):
        self.pool.release_connection(self, destroy)

    def reply(self, api_reply_index):
        value = socket.htonl(api_reply_index)
        try:
            self.socket.sendall(struct.pack('I', value))
        except:
            self.release(True)
            raise NetworkException("Unable to send API reply")

    def requires_cs_negotiation(self):
        try:
            if self.account.client_server_negotiation == REQUEST_NEGOTIATION:
                return True
        except AttributeError:
            return False
        return False

    def ssl_startup(self):
        # Get encryption settings from client environment
        host = self.account.host
        algo = self.account.encryption_algorithm
        key_size = self.account.encryption_key_size
        hash_rounds = self.account.encryption_num_hash_rounds
        salt_size = self.account.encryption_salt_size

        # Get or create SSL context
        try:
            context = self.account.ssl_context
        except AttributeError:
            CA_file = getattr(self.account, 'ssl_ca_certificate_file', None)
            verify_server_mode = getattr(self.account,'ssl_verify_server', 'hostname')
            if verify_server_mode == 'none':
                CA_file = None
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=CA_file)
            if CA_file is None:
                context.check_hostname = False
                context.verify_mode = 0  # VERIFY_NONE

        # Wrap socket with context
        wrapped_socket = context.wrap_socket(self.socket, server_hostname=host)

        # Initial SSL handshake
        wrapped_socket.do_handshake()

        # Generate key (shared secret)
        key = os.urandom(self.account.encryption_key_size)

        # Send header-only message with client side encryption settings
        packed_header = iRODSMessage.pack_header(algo,
                                                 key_size,
                                                 salt_size,
                                                 hash_rounds,
                                                 0)
        wrapped_socket.sendall(packed_header)

        # Send shared secret
        packed_header = iRODSMessage.pack_header('SHARED_SECRET',
                                                 key_size,
                                                 0,
                                                 0,
                                                 0)
        wrapped_socket.sendall(packed_header + key)

        # Use SSL socket from now on
        self.socket = wrapped_socket

    def _connect(self):
        address = (self.account.host, self.account.port)
        timeout = self.pool.connection_timeout

        try:
            s = socket.create_connection(address, timeout)
            self._disconnected = False
        except socket.error:
            raise NetworkException(
                "Could not connect to specified host and port: " +
                "{}:{}".format(*address))

        self.socket = s

        main_message = StartupPack(
            (self.account.proxy_user, self.account.proxy_zone),
            (self.account.client_user, self.account.client_zone),
            self.pool.application_name
        )

        # No client-server negotiation
        if not self.requires_cs_negotiation():

            # Send startup pack without negotiation request
            msg = iRODSMessage(msg_type='RODS_CONNECT', msg=main_message)
            self.send(msg)

            # Server responds with version
            version_msg = self.recv()

            # Done
            return version_msg.get_main_message(VersionResponse)

        # Get client negotiation policy
        client_policy = getattr(self.account, 'client_server_policy', REQUIRE_TCP)

        # Sanity check
        validate_policy(client_policy)

        # Send startup pack with negotiation request
        main_message.option = '{};{}'.format(main_message.option, REQUEST_NEGOTIATION)
        msg = iRODSMessage(msg_type='RODS_CONNECT', msg=main_message)
        self.send(msg)

        # Server responds with its own negotiation policy
        cs_neg_msg = self.recv()
        response = cs_neg_msg.get_main_message(ClientServerNegotiation)
        server_policy = response.result

        # Perform the negotiation
        neg_result, status = perform_negotiation(client_policy=client_policy,
                                                 server_policy=server_policy)

        # Send negotiation result to server
        client_neg_response = ClientServerNegotiation(
            status=status,
            result='{}={};'.format(CS_NEG_RESULT_KW, neg_result)
        )
        msg = iRODSMessage(msg_type='RODS_CS_NEG_T', msg=client_neg_response)
        self.send(msg)

        # If negotiation failed we're done
        if neg_result == FAILURE:
            self.disconnect()
            raise NetworkException("Client-Server negotiation failure: {},{}".format(client_policy, server_policy))

        # Server responds with version
        version_msg = self.recv()

        if neg_result == USE_SSL:
            self.ssl_startup()

        return version_msg.get_main_message(VersionResponse)

    def disconnect(self):
        disconnect_msg = iRODSMessage(msg_type='RODS_DISCONNECT')
        self.send(disconnect_msg)
        try:
            # SSL shutdown handshake
            self.socket = self.socket.unwrap()
        except AttributeError:
            pass
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        self.socket = None
        self._disconnected = True

    def recvall(self, n):
        # Helper function to recv n bytes or return None if EOF is hit
        data = b''
        while len(data) < n:
            packet = self.socket.recv(n - len(data))
            if not packet:
                return None
            data += packet
        return data

    def init_sec_context(self):
        import gssapi

        # AUTHORIZATION MECHANISM
        gsi_mech = gssapi.raw.OID.from_int_seq(GSI_OID)

        # SERVER NAME
        server_name = gssapi.Name(self.account.server_dn)
        server_name.canonicalize(gsi_mech)

        # CLIENT CONTEXT
        self.client_ctx = gssapi.SecurityContext(
            name=server_name,
            mech=gsi_mech,
            flags=[2, 4],
            usage='initiate')

    def send_gsi_token(self, server_token=None):

        # CLIENT TOKEN
        if server_token is None:
            client_token = self.client_ctx.step()
        else:
            client_token = self.client_ctx.step(server_token)
        logger.debug("[GSI handshake] Client: sent a new token")

        # SEND IT TO SERVER
        self.reply(len(client_token))
        self.socket.sendall(client_token)

    def receive_gsi_token(self):

        # Receive client token from iRODS
        data = self.socket.recv(4)
        value = struct.unpack("I", bytearray(data))
        token_len = socket.ntohl(value[0])
        server_token = self.recvall(token_len)
        logger.debug("[GSI handshake] Server: received a new token")

        return server_token

    def handshake(self, target):
        """
        This GSS API context based on GSI was obtained combining 2 sources:
    https://pythonhosted.org/gssapi/basic-tutorial.html
    https://github.com/irods/irods_auth_plugin_gsi/blob/master/gsi/libgsi.cpp
        """

        self.init_sec_context()

        # Go, into the loop
        self.send_gsi_token()

        while not (self.client_ctx.complete):

            server_token = self.receive_gsi_token()

            self.send_gsi_token(server_token)

        logger.debug("[GSI Handshake] completed")

    def gsi_client_auth_request(self):

        # Request for authentication with GSI on current user

        message_body = PluginAuthMessage(
            auth_scheme_=GSI_AUTH_PLUGIN,
            context_='%s=%s' % (AUTH_USER_KEY, self.account.client_user)
        )
        # GSI = 1201
# https://github.com/irods/irods/blob/master/lib/api/include/apiNumber.h#L158
        auth_req = iRODSMessage(
            msg_type='RODS_API_REQ', msg=message_body, int_info=1201)
        self.send(auth_req)
        # Getting the challenge message
        self.recv()

        # This receive an empty message for confirmation... To check:
        # challenge_msg = self.recv()

    def gsi_client_auth_response(self):

        message = '%s=%s' % (AUTH_SCHEME_KEY, GSI_AUTH_SCHEME)
        # IMPORTANT! padding
        len_diff = RESPONSE_LEN - len(message)
        message += "\0" * len_diff

        # mimic gsi_auth_client_response
        gsi_msg = AuthResponse(
            response=message,
            username=self.account.proxy_user + '#' + self.account.proxy_zone
        )
        gsi_request = iRODSMessage(
            msg_type='RODS_API_REQ', int_info=704, msg=gsi_msg)
        self.send(gsi_request)
        self.recv()
        # auth_response = self.recv()

    def _login_gsi(self):
        # Send iRODS server a message to request GSI authentication
        self.gsi_client_auth_request()

        # Create a context handshaking GSI credentials
        # Note: this can work only if you export GSI certificates
        # as shell environment variables (X509_etc.)
        self.handshake(self.account.host)

        # Complete the protocol
        self.gsi_client_auth_response()

        logger.info("GSI authorization validated")

    def _login_pam(self):

        time_to_live_in_seconds = 60

        ctx_user = '%s=%s' % (AUTH_USER_KEY, self.account.client_user)
        ctx_pwd = '%s=%s' % (AUTH_PWD_KEY, self.account.password)
        ctx_ttl = '%s=%s' % (AUTH_TTL_KEY, str(time_to_live_in_seconds))

        ctx = ";".join([ctx_user, ctx_pwd, ctx_ttl])

        if type(self.socket) is socket.socket:
            if getattr(self,'DISALLOWING_PAM_PLAINTEXT',True):
                raise PlainTextPAMPasswordError

        Pam_Long_Tokens = (ALLOW_PAM_LONG_TOKENS and (len(ctx) >= MAX_NAME_LEN))

        if Pam_Long_Tokens:

            message_body = PamAuthRequest(
                pamUser=self.account.client_user,
                pamPassword=self.account.password,
                timeToLive=time_to_live_in_seconds)
        else:

            message_body = PluginAuthMessage(
                auth_scheme_ = PAM_AUTH_SCHEME,
                context_ = ctx)

        auth_req = iRODSMessage(
            msg_type='RODS_API_REQ',
            msg=message_body,
            int_info=(725 if Pam_Long_Tokens else 1201)
        )

        self.send(auth_req)
        # Getting the new password
        output_message = self.recv()

        Pam_Response_Class = (PamAuthRequestOut if Pam_Long_Tokens
                         else AuthPluginOut)

        auth_out = output_message.get_main_message( Pam_Response_Class )

        self.disconnect()
        self._connect()

        if hasattr(self.account,'store_pw'):
            drop = self.account.store_pw
            if type(drop) is list:
                drop[:] = [ auth_out.result_ ]

        self._login_native(password=auth_out.result_)

        logger.info("PAM authorization validated")

    def read_file(self, desc, size=-1, buffer=None):
        if size < 0:
            size = len(buffer)
        elif buffer is not None:
            size = min(size, len(buffer))

        message_body = OpenedDataObjRequest(
            l1descInx=desc,
            len=size,
            whence=0,
            oprType=0,
            offset=0,
            bytesWritten=0,
            KeyValPair_PI=StringStringMap()
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_READ_AN'])

        logger.debug(desc)
        self.send(message)
        if buffer is None:
            response = self.recv()
        else:
            response = self.recv_into(buffer)

        return response.bs

    def _login_native(self, password=None):

        # Default case, PAM login will send a new password
        if password is None:
            password = self.account.password

        # authenticate
        auth_req = iRODSMessage(msg_type='RODS_API_REQ', int_info=703)
        self.send(auth_req)

        # challenge
        challenge_msg = self.recv()
        logger.debug(challenge_msg.msg)
        challenge = challenge_msg.get_main_message(AuthChallenge).challenge

        # one "session" signature per connection
        # see https://github.com/irods/irods/blob/4.2.1/plugins/auth/native/libnative.cpp#L137
        # and https://github.com/irods/irods/blob/4.2.1/lib/core/src/clientLogin.cpp#L38-L60
        if six.PY2:
            self._client_signature = "".join("{:02x}".format(ord(c)) for c in challenge[:16])
        else:
            self._client_signature = "".join("{:02x}".format(c) for c in challenge[:16])

        if six.PY3:
            challenge = challenge.strip()
            padded_pwd = struct.pack(
                "%ds" % MAX_PASSWORD_LENGTH, password.encode(
                    'utf-8').strip())
        else:
            padded_pwd = struct.pack(
                "%ds" % MAX_PASSWORD_LENGTH, password)

        m = hashlib.md5()
        m.update(challenge)
        m.update(padded_pwd)
        encoded_pwd = m.digest()

        if six.PY2:
            encoded_pwd = encoded_pwd.replace('\x00', '\x01')
        elif b'\x00' in encoded_pwd:
            encoded_pwd_array = bytearray(encoded_pwd)
            encoded_pwd = bytes(encoded_pwd_array.replace(b'\x00', b'\x01'))

        pwd_msg = AuthResponse(
            response=encoded_pwd, username=self.account.proxy_user)
        pwd_request = iRODSMessage(
            msg_type='RODS_API_REQ', int_info=704, msg=pwd_msg)
        self.send(pwd_request)
        self.recv()

    def write_file(self, desc, string):
        message_body = OpenedDataObjRequest(
            l1descInx=desc,
            len=len(string),
            whence=0,
            oprType=0,
            offset=0,
            bytesWritten=0,
            KeyValPair_PI=StringStringMap()
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               bs=string,
                               int_info=api_number['DATA_OBJ_WRITE_AN'])
        self.send(message)
        response = self.recv()
        return response.int_info

    def seek_file(self, desc, offset, whence):
        message_body = OpenedDataObjRequest(
            l1descInx=desc,
            len=0,
            whence=whence,
            oprType=0,
            offset=offset,
            bytesWritten=0,
            KeyValPair_PI=StringStringMap()
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_LSEEK_AN'])

        self.send(message)
        response = self.recv()
        offset = response.get_main_message(FileSeekResponse).offset
        return offset

    def close_file(self, desc, **options):
        message_body = OpenedDataObjRequest(
            l1descInx=desc,
            len=0,
            whence=0,
            oprType=0,
            offset=0,
            bytesWritten=0,
            KeyValPair_PI=StringStringMap(options)
        )
        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number['DATA_OBJ_CLOSE_AN'])

        self.send(message)
        self.recv()

    def temp_password(self):
        request = iRODSMessage("RODS_API_REQ", msg=None,
                               int_info=api_number['GET_TEMP_PASSWORD_AN'])

        # Send and receive request
        self.send(request)
        response = self.recv()
        logger.debug(response.int_info)

        # Convert and return answer
        msg = response.get_main_message(GetTempPasswordOut)
        return obf.create_temp_password(msg.stringToHashWith, self.account.password)
